"""Define custom proximal operators for use with copt package."""
from __future__ import absolute_import, division, print_function

import numpy as np

from itertools import compress

__all__ = ["SparseGroupL1"]


def _soft_threshold(z, alpha):
    r"""Apply the element-wise soft thresholding operator.

    The soft-thresholding operator is

    .. math::
        S(z_i, alpha) = \begin{cases}
            z_i + alpha & z_i < -T \\
            0 & -alpha \le z_i \le alpha \\
            z_i - alpha & alpha < z_i
        \end{cases}
        (
        \text{prox}_{\sigma P_2} \circ \text{prox}_{\sigma P_1}
        \right)(u)


    Parameters
    ----------
    z : array-like
        Input array

    alpha : float
        threshold value

    Returns
    -------
    np.ndarray
        Element-wise soft-thresholded array
    """  # noqa: W605
    return np.fmax(z - alpha, 0) - np.fmax(-z - alpha, 0)


class SparseGroupL1(object):
    r"""Sparse group lasso penalty class for use with openopt/copt package.

    Implements the sparse group lasso penalty [1]_

    .. math::
        (1 - \rho) * \alpha \displaystyle \sum_{g \in G} || \beta_g ||_2
        + \rho * \alpha || \beta ||_1

    where :math:`G` is a partition of the features into groups.

    Parameters
    ----------
    l1_ratio : float
        Combination between group lasso and lasso. l1_ratio = 0 gives the
        group lasso and l1_ratio = 1 gives the lasso. ``l1_ratio``
        corresponds to `\rho` in the equation above

    alpha : float
        Regularization parameter, overall strength of regularization.

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    bias_index : int or None, default=None
        If None, regularize all coefficients. Otherwise, this is the index
        of the bias (i.e. intercept) feature, which should not be regularized.
        Exclude this index from the penalty term. And force the proximal
        operator for this index to return the result of the identity function.

    References
    ----------
    .. [1]  Noah Simon, Jerome Friedman, Trevor Hastie & Robert Tibshirani,
        "A Sparse-Group Lasso," Journal of Computational and Graphical
        Statistics, vol. 22:2, pp. 231-245, 2012
        DOI: 10.1080/10618600.2012.681250
    """  # noqa: W605

    def __init__(
        self, l1_ratio, alpha, groups, bias_index=None, scale_l2_by="group_length"
    ):
        self.l1_ratio = l1_ratio
        self.alpha = alpha
        self.groups = groups
        self.feature2group_map = np.concatenate(
            [
                [grp_idx] * feature_indices.size
                for grp_idx, feature_indices in enumerate(groups)
            ]
        )

        n_features = np.unique(np.concatenate(groups)).size
        if bias_index is not None:
            n_features += 1

        self.group_masks = np.full((len(groups), n_features), 0)
        for row, grp in zip(self.group_masks, groups):
            row[grp] = 1

        if scale_l2_by == "group_length":
            self.group_scale = np.sqrt([grp.size for grp in groups])
        else:
            self.group_scale = np.array([1.0 for grp in groups])

        self.bias_index = bias_index

    def __call__(self, x):
        """Return the Sparse L1 penalty."""
        penalty = (
            (1.0 - self.l1_ratio)
            * self.alpha
            * self.group_scale
            * np.sum([np.linalg.norm(x[g]) for g in self.groups])
        )

        ind = np.ones(len(x), bool)
        if self.bias_index is not None:
            ind[self.bias_index] = False

        penalty += self.l1_ratio * self.alpha * np.abs(x[ind]).sum()
        return penalty

    def prox(self, x, step_size):
        r"""Return the proximal operator of the sparse group lasso penalty.

        For the sparse group lasso, we can decompose the penalty into

        .. math::
            P(\beta) = P_1(\beta) + P_2(\beta)

        where :math:`P_2 = \rho \alpha || \beta ||_1` is the lasso
        penalty and :math:`P_1 = (1 - \rho) \alpha \displaystyle
        \sum_{g \in G} || \beta_g ||_2` is the group lasso penalty.

        Then the proximal operator is given by

        .. math::
            \text{prox}_{\sigma P_1 + \sigma P_2} (u) = \left(
            \text{prox}_{\sigma P_2} \circ \text{prox}_{\sigma P_1}
            \right)(u)

        where :math:`sigma` is a step size

        Parameters
        ----------
        x : np.ndarray
            Argument for proximal operator.

        step_size : float
            Step size for proximal operator

        Returns
        -------
        np.ndarray
            proximal operator of sparse group lasso penalty evaluated on
            input `x` with step size `step_size`
        """  # noqa: W605
        l1_prox = _soft_threshold(x, self.l1_ratio * self.alpha * step_size)
        out = l1_prox.copy()

        if self.bias_index is not None:
            out[self.bias_index] = x[self.bias_index]

        norms = np.sqrt(self.group_masks.dot(l1_prox ** 2)) / self.group_scale

        norm_mask = norms > (1.0 - self.l1_ratio) * self.alpha * step_size
        all_norm = all(norm_mask)
        if not all_norm:
            idx_true = np.array([], dtype=int)
        else:
            idx_true = np.concatenate(list(compress(self.groups, norm_mask)))

        if all_norm:
            idx_false = np.array([], dtype=int)
        else:
            idx_false = np.concatenate(
                list(compress(self.groups, np.logical_not(norm_mask)))
            )

        out[idx_true] -= (
            step_size
            * (1.0 - self.l1_ratio)
            * self.alpha
            * out[idx_true]
            / norms[self.feature2group_map][idx_true]
        )
        out[idx_false] = 0.0

        return out



import numpy as np
import pytest

from groupyr.transform import isiterable, GroupExtractor
from groupyr.transform import GroupRemover, GroupShuffler
from sklearn.utils.estimator_checks import check_estimator
from sklearn.utils import shuffle as util_shuffle


def test_isiterable():
    assert isiterable(range(10))  # nosec
    assert not isiterable(5)  # nosec
    assert isiterable(np.arange(10))  # nosec


def test_GroupExtractor():
    X = np.array([list(range(10))] * 10)
    groups = [
        np.array([0, 1, 2]),
        np.array([3, 4, 5]),
        np.array([6, 7, 8]),
        np.array([9]),
    ]
    group_names = ["one", "two", "three", "four"]
    group_tuple_names = [
        ("one", "alpha"),
        ("one", "beta"),
        ("two", "alpha"),
        ("three", "beta"),
    ]

    select = 2
    ge = GroupExtractor(groups=groups, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, groups[select]], X_tr)  # nosec

    select = "two"
    idx = np.concatenate([grp for grp, t in zip(groups, group_names) if t == select])
    ge = GroupExtractor(groups=groups, group_names=group_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = ["two", "three"]
    idx = np.concatenate([grp for grp, t in zip(groups, group_names) if t in select])
    ge = GroupExtractor(groups=groups, group_names=group_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = "one"
    idx = np.concatenate(
        [grp for grp, t in zip(groups, group_tuple_names) if t[0] == select]
    )
    ge = GroupExtractor(groups=groups, group_names=group_tuple_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = ["alpha", "three"]
    idx = np.concatenate(
        [
            grp
            for grp, t in zip(groups, group_tuple_names)
            if any([r in t for r in select])
        ]
    )
    ge = GroupExtractor(groups=groups, group_names=group_tuple_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = [0, 3]
    ge = GroupExtractor(groups=groups, select=select)
    X_tr = ge.fit_transform(X)
    idx = np.concatenate([groups[e] for e in select])
    assert np.allclose(X[:, idx], X_tr)  # nosec

    ge = GroupExtractor()
    X_tr = ge.fit_transform(X)
    assert np.allclose(X_tr, X)  # nosec

    with pytest.raises(ValueError):
        GroupExtractor(select="error").fit_transform(X)


def test_GroupRemover():
    X = np.array([list(range(10))] * 10)
    groups = [
        np.array([0, 1, 2]),
        np.array([3, 4, 5]),
        np.array([6, 7, 8]),
        np.array([9]),
    ]
    group_names = ["one", "two", "three", "four"]
    group_tuple_names = [
        ("one", "alpha"),
        ("one", "beta"),
        ("two", "alpha"),
        ("three", "beta"),
    ]

    select = 2
    idx = np.concatenate([grp for idx, grp in enumerate(groups) if idx != select])
    ge = GroupRemover(groups=groups, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = "two"
    idx = np.concatenate([grp for grp, t in zip(groups, group_names) if t != select])
    ge = GroupRemover(groups=groups, group_names=group_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = ["two", "three"]
    idx = np.concatenate(
        [grp for grp, t in zip(groups, group_names) if t not in select]
    )
    ge = GroupRemover(groups=groups, group_names=group_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = "one"
    idx = np.concatenate(
        [grp for grp, t in zip(groups, group_tuple_names) if t[0] != select]
    )
    ge = GroupRemover(groups=groups, group_names=group_tuple_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = ["alpha", "three"]
    idx = np.concatenate(
        [
            grp
            for grp, t in zip(groups, group_tuple_names)
            if all([r not in t for r in select])
        ]
    )
    ge = GroupRemover(groups=groups, group_names=group_tuple_names, select=select)
    X_tr = ge.fit_transform(X)
    assert np.allclose(X[:, idx], X_tr)  # nosec

    select = [0, 3]
    ge = GroupRemover(groups=groups, select=select)
    X_tr = ge.fit_transform(X)
    idx = np.concatenate([grp for idx, grp in enumerate(groups) if idx not in select])
    assert np.allclose(X[:, idx], X_tr)  # nosec

    ge = GroupRemover()
    X_tr = ge.fit_transform(X)
    assert np.allclose(X_tr, X)  # nosec

    with pytest.raises(ValueError):
        GroupRemover(select=object()).fit_transform(X)

    with pytest.raises(ValueError):
        GroupRemover(group_names=group_names).fit_transform(X)

    with pytest.raises(ValueError):
        GroupRemover(groups=groups, group_names=["a", "b"]).fit_transform(X)

    with pytest.raises(ValueError):
        GroupRemover(groups=groups, select="b").fit_transform(X)

    with pytest.raises(ValueError):
        GroupRemover(groups=groups, select=["a", "b"]).fit_transform(X)

    with pytest.raises(TypeError):
        GroupRemover(groups=groups, group_names=0).fit_transform(X)


def assert_shuffle_match(
    X, X_tr, groups=None, group_names=None, select=None, random_state=0
):
    ge = GroupExtractor(select=select, groups=groups, group_names=group_names)
    shuffled = {"orig": ge.fit_transform(X), "tran": ge.fit_transform(X_tr)}
    gr = GroupRemover(select=select, groups=groups, group_names=group_names)
    same = {"orig": gr.fit_transform(X), "tran": gr.fit_transform(X_tr)}
    assert np.allclose(same["orig"], same["tran"])
    assert np.allclose(
        util_shuffle(shuffled["orig"], random_state=random_state), shuffled["tran"]
    )


def test_GroupShuffler():
    X = np.arange(100).reshape(10, 10)
    groups = [
        np.array([0, 1, 2]),
        np.array([3, 4, 5]),
        np.array([6, 7, 8]),
        np.array([9]),
    ]
    group_names = ["one", "two", "three", "four"]
    group_tuple_names = [
        ("one", "alpha"),
        ("one", "beta"),
        ("two", "alpha"),
        ("three", "beta"),
    ]

    select = 2
    gs = GroupShuffler(groups=groups, select=select, random_state=0)
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(X, X_tr, groups=groups, select=select, random_state=0)

    select = "two"
    gs = GroupShuffler(
        groups=groups, group_names=group_names, select=select, random_state=0
    )
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(
        X, X_tr, groups=groups, group_names=group_names, select=select, random_state=0
    )

    select = ["two", "three"]
    gs = GroupShuffler(
        groups=groups, group_names=group_names, select=select, random_state=0
    )
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(
        X, X_tr, groups=groups, group_names=group_names, select=select, random_state=0
    )

    select = "one"
    gs = GroupShuffler(
        groups=groups, group_names=group_tuple_names, select=select, random_state=0
    )
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(
        X,
        X_tr,
        groups=groups,
        group_names=group_tuple_names,
        select=select,
        random_state=0,
    )

    select = ["alpha", "three"]
    gs = GroupShuffler(
        groups=groups, group_names=group_tuple_names, select=select, random_state=0
    )
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(
        X,
        X_tr,
        groups=groups,
        group_names=group_tuple_names,
        select=select,
        random_state=0,
    )

    select = [0, 3]
    gs = GroupShuffler(groups=groups, select=select, random_state=0)
    X_tr = gs.fit_transform(X)
    assert_shuffle_match(X, X_tr, groups=groups, select=select, random_state=0)

    gs = GroupShuffler()
    X_tr = gs.fit_transform(X)
    assert np.allclose(X_tr, X)  # nosec


@pytest.mark.parametrize("Transformer", [GroupExtractor, GroupRemover, GroupShuffler])
def test_all_estimators(Transformer):
    return check_estimator(Transformer())



"""
=======================================
Visualizing the SGL regularization path
=======================================

Computes SGL path using the function ``sgl_path`` along the regularization parameter
on a synthetic dataset.  Each color represents a different feature group in the
coefficient vector, and this is displayed as a function of the regularization
parameter.

When alpha is very large, the regularization effect dominates and the
coefficients tend to zero.  As alpha tends toward zero the solution tends
towards the ordinary least squares.

When a group of covariates goes to zero (indicated by the dashed vertical line),
this relieves the regularization penalty on the other groups, creating "jumps"
in the magnitudes of the other covariate groups.

In practise it is necessary to tune alpha (and l1_ratio) to best fit the
training data.
"""
import groupyr as gpr
import matplotlib.pyplot as plt
import numpy as np


X, y, groups, coef = gpr.datasets.make_group_regression(
    n_samples=400,
    n_groups=10,
    n_informative_groups=3,
    n_features_per_group=10,
    n_informative_per_group=3,
    noise=200,
    coef=True,
    random_state=10,
    shuffle=True,
)

path_coefs, path_alphas, path_iters = gpr.sgl_path(
    X, y, l1_ratio=0.5, groups=groups, eps=0.001, n_alphas=200, max_ier=1000, tol=1e-3
)

group_means = np.array([np.linalg.norm(path_coefs[grp], axis=0) for grp in groups])

zero_idx = np.array(
    [np.max(np.where(group_means[i] == 0)[0]) for i in range(len(groups))]
)

fig, ax = plt.subplots(2, 1, figsize=(8, 10), sharex=True)

cmap = plt.get_cmap("tab10")

for grp, color, z_idx in zip(groups, cmap.colors, zero_idx):
    _ = ax[0].semilogx(
        path_alphas, np.abs(path_coefs[grp][:-1].transpose()), color=color
    )

    _ = ax[0].axvline(path_alphas[z_idx], ls=":", color=color)
    _ = ax[1].axvline(path_alphas[z_idx], ls=":", color=color)

_ = ax[1].semilogx(path_alphas, group_means.transpose())

_ = ax[1].set_xlabel(r"$\log(\alpha)$", fontsize=16)
_ = ax[0].set_ylabel(r"$\left| \hat{\beta} \right|$", fontsize=16)
_ = ax[1].set_ylabel(
    r"$\left| \left| \hat{\beta}^{(\ell)} \right| \right|_2$", fontsize=16
)

_ = ax[0].set_title(r"SGL ($\lambda=0.5$) regularization path", fontsize=16)



import numpy as np
import pytest

from groupyr._prox import SparseGroupL1

proximal_penalties = [
    (SparseGroupL1(0.5, 1.0, groups=[np.arange(16)]), 16),
    (SparseGroupL1(0.5, 1.0, groups=[np.arange(16)], scale_l2_by=None), 16),
    (SparseGroupL1(0.5, 1.0, groups=[np.arange(16)], bias_index=16), 17),
]


@pytest.mark.parametrize("penalty, n_features", proximal_penalties)
def test_three_inequality(penalty, n_features):
    """Test the prox using the three point inequality
    The three-point inequality is described e.g., in Lemma 1.4
    in "Gradient-Based Algorithms with Applications to Signal
    Recovery Problems", Amir Beck and Marc Teboulle
    """
    for _ in range(10):
        z = np.random.randn(n_features)
        u = np.random.randn(n_features)
        xi = penalty.prox(z, 1.0)

        lhs = 2 * (penalty(xi) - penalty(u))
        rhs = (
            np.linalg.norm(u - z) ** 2
            - np.linalg.norm(u - xi) ** 2
            - np.linalg.norm(xi - z) ** 2
        )

        assert lhs <= rhs, penalty  # nosec



"""Generate samples of synthetic data sets."""
import numpy as np

from sklearn.datasets import make_classification, make_regression
from sklearn.utils import check_random_state
from sklearn.utils import shuffle as util_shuffle

__all__ = ["make_group_classification", "make_group_regression"]


def make_group_classification(
    n_samples=100,
    n_groups=20,
    n_informative_groups=2,
    n_features_per_group=20,
    n_informative_per_group=2,
    n_redundant_per_group=2,
    n_repeated_per_group=0,
    n_classes=2,
    n_clusters_per_class=2,
    weights=None,
    flip_y=0.01,
    class_sep=1.0,
    hypercube=True,
    shift=0.0,
    scale=1.0,
    shuffle=True,
    useful_indices=False,
    random_state=None,
):
    """Generate a random n-class sparse group classification problem.

    This initially creates clusters of points normally distributed (std=1)
    about vertices of an ``n_informative``-dimensional hypercube with sides of
    length ``2*class_sep`` and assigns an equal number of clusters to each
    class. It introduces interdependence between these features and adds
    various types of further noise to the data.

    Prior to shuffling, ``X`` stacks a number of these primary "informative"
    features, "redundant" linear combinations of these, "repeated" duplicates
    of sampled features, and arbitrary noise for and remaining features.
    This method uses sklearn.datasets.make_classification to construct a
    giant unshuffled classification problem of size
    ``n_groups * n_features_per_group`` and then distributes the returned
    features to each group. It then optionally shuffles each group.

    Parameters
    ----------
    n_samples : int, optional (default=100)
        The number of samples.

    n_groups : int, optional (default=10)
        The number of feature groups.

    n_informative_groups : int, optional (default=2)
        The total number of informative groups. All other groups will be
        just noise.

    n_features_per_group : int, optional (default=20)
        The total number of features_per_group. These comprise `n_informative`
        informative features, `n_redundant` redundant features, `n_repeated`
        duplicated features and `n_features-n_informative-n_redundant-
        n_repeated` useless features drawn at random.

    n_informative_per_group : int, optional (default=2)
        The number of informative features_per_group. Each class is composed
        of a number of gaussian clusters each located around the vertices of a
        hypercube in a subspace of dimension `n_informative_per_group`. For
        each cluster, informative features are drawn independently from
        N(0, 1) and then randomly linearly combined within each cluster in
        order to add covariance. The clusters are then placed on the vertices
        of the hypercube.

    n_redundant_per_group : int, optional (default=2)
        The number of redundant features per group. These features are
        generated as random linear combinations of the informative features.

    n_repeated_per_group : int, optional (default=0)
        The number of duplicated features per group, drawn randomly from the
        informative and the redundant features.

    n_classes : int, optional (default=2)
        The number of classes (or labels) of the classification problem.

    n_clusters_per_class : int, optional (default=2)
        The number of clusters per class.

    weights : list of floats or None (default=None)
        The proportions of samples assigned to each class. If None, then
        classes are balanced. Note that if `len(weights) == n_classes - 1`,
        then the last class weight is automatically inferred.
        More than `n_samples` samples may be returned if the sum of `weights`
        exceeds 1.

    flip_y : float, optional (default=0.01)
        The fraction of samples whose class are randomly exchanged. Larger
        values introduce noise in the labels and make the classification
        task harder.

    class_sep : float, optional (default=1.0)
        The factor multiplying the hypercube size.  Larger values spread
        out the clusters/classes and make the classification task easier.

    hypercube : boolean, optional (default=True)
        If True, the clusters are put on the vertices of a hypercube. If
        False, the clusters are put on the vertices of a random polytope.

    shift : float, array of shape [n_features] or None, optional (default=0.0)
        Shift features by the specified value. If None, then features
        are shifted by a random value drawn in [-class_sep, class_sep].

    scale : float, array of shape [n_features] or None, optional (default=1.0)
        Multiply features by the specified value. If None, then features
        are scaled by a random value drawn in [1, 100]. Note that scaling
        happens after shifting.

    shuffle : boolean, optional (default=True)
        Shuffle the samples and the features.

    useful_indices : boolean, optional (default=False)
        If True, a boolean array indicating useful features is returned

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.

    y : array of shape [n_samples]
        The integer labels for class membership of each sample.

    groups : list of arrays
        Each element is an array of feature indices that belong to that group

    indices : array of shape [n_features]
        A boolean array indicating which features are useful. Returned only
        if `useful_indices` is True.

    Notes
    -----
    The algorithm is adapted from Guyon [1] and was designed to generate
    the "Madelon" dataset.

    References
    ----------
    .. [1] I. Guyon, "Design of experiments for the NIPS 2003 variable
           selection benchmark", 2003.

    See Also
    --------
    sklearn.datasets.make_classification: non-group-sparse version
    sklearn.datasets.make_blobs: simplified variant
    sklearn.datasets.make_multilabel_classification: unrelated generator for multilabel tasks
    """
    generator = check_random_state(random_state)

    total_features = n_groups * n_features_per_group
    total_informative = n_informative_groups * n_informative_per_group
    total_redundant = n_informative_groups * n_redundant_per_group
    total_repeated = n_informative_groups * n_repeated_per_group

    # Count features, clusters and samples
    if (
        n_informative_per_group + n_redundant_per_group + n_repeated_per_group
        > n_features_per_group
    ):
        raise ValueError(
            "Number of informative, redundant and repeated features per group"
            " must sum to less than the number of total features per group."
        )

    # Generate a big classification problem for the total number of features
    # The `shuffle` argument is False so that the feature matrix X has
    # features stacked in the order: informative, redundant, repeated, useless
    X, y = make_classification(
        n_samples=n_samples,
        n_features=total_features,
        n_informative=total_informative,
        n_redundant=total_redundant,
        n_repeated=total_repeated,
        n_classes=n_classes,
        n_clusters_per_class=n_clusters_per_class,
        weights=weights,
        flip_y=flip_y,
        class_sep=class_sep,
        hypercube=hypercube,
        shift=shift,
        scale=scale,
        shuffle=False,
        random_state=generator,
    )

    total_useful = total_informative + total_redundant + total_repeated
    idx = np.arange(total_features) < total_useful

    # Evenly distribute the first `n_informative_groups * n_features_per_group`
    # features into the first `n_informative_groups` groups
    n_info_grp_features = n_informative_groups * n_features_per_group
    idx_range = np.arange(n_info_grp_features)

    idx_map_consolidated_2_grouped = (
        np.concatenate(
            [np.arange(0, n_info_grp_features, n_informative_groups)]
            * n_informative_groups
        )
        + idx_range // n_features_per_group
    )

    X = np.concatenate(
        [X[:, idx_map_consolidated_2_grouped], X[:, n_info_grp_features:]], axis=1
    )

    if useful_indices:
        idx = np.concatenate(
            [idx[idx_map_consolidated_2_grouped], idx[n_info_grp_features:]]
        )

    if shuffle:
        # Randomly permute samples
        X, y = util_shuffle(X, y, random_state=generator)

        # Permute the groups, maintaining the order within them group_idx_map
        # maps feature indices to group indices. The group order is random
        # but all features in a single group are adjacent
        group_idx_map = np.concatenate(
            [
                np.ones(n_features_per_group, dtype=np.int32) * i
                for i in generator.choice(
                    np.arange(n_groups), size=n_groups, replace=False
                )
            ]
        )

        permute_group_map = (
            np.concatenate(
                [
                    generator.choice(
                        np.arange(n_features_per_group),
                        size=n_features_per_group,
                        replace=False,
                    )
                    for _ in range(n_groups)
                ]
            )
            + group_idx_map * n_features_per_group
        )

        X = X[:, permute_group_map]

        if useful_indices:
            idx = idx[permute_group_map]
    else:
        group_idx_map = np.concatenate(
            [np.ones(n_features_per_group, dtype=np.int32) * i for i in range(n_groups)]
        )

    groups = [np.where(group_idx_map == idx)[0] for idx in range(n_groups)]

    X = np.ascontiguousarray(X)
    if useful_indices:
        return X, y, groups, idx
    else:
        return X, y, groups


def make_group_regression(
    n_samples=100,
    n_groups=20,
    n_informative_groups=5,
    n_features_per_group=20,
    n_informative_per_group=5,
    effective_rank=None,
    noise=0.0,
    shift=0.0,
    scale=1.0,
    shuffle=False,
    coef=False,
    random_state=None,
):
    """Generate a sparse group regression problem.

    Prior to shuffling, ``X`` stacks a number of these primary "informative"
    features, and arbitrary noise for and remaining features.
    This method uses sklearn.datasets.make_regression to construct a
    giant unshuffled regression problem of size
    ``n_groups * n_features_per_group`` and then distributes the returned
    features to each group. It then optionally shuffles each group.

    Parameters
    ----------
    n_samples : int, optional (default=100)
        The number of samples.

    n_groups : int, optional (default=10)
        The number of feature groups.

    n_informative_groups : int, optional (default=2)
        The total number of informative groups. All other groups will be
        just noise.

    n_features_per_group : int, optional (default=20)
        The total number of features_per_group. These comprise `n_informative`
        informative features, and `n_features-n_informative` useless
        features drawn at random.

    n_informative_per_group : int, optional (default=2)
        The number of informative features_per_group that have a
        non-zero regression coefficient.

    effective_rank : int or None, optional (default=None)
        If not None, provides the number of singular vectors to explain the
        input data.

    noise : float, optional (default=0.0)
         The standard deviation of the gaussian noise applied to the output.

    shuffle : boolean, optional (default=False)
        Shuffle the samples and the features.

    coef : boolean, optional (default=False)
        If True, returns coefficient values used to generate samples via
        sklearn.datasets.make_regression.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    X : array of shape [n_samples, n_features]
        The generated samples.

    y : array of shape [n_samples]
        The integer labels for class membership of each sample.

    groups : list of arrays
        Each element is an array of feature indices that belong to that group

    coef : array of shape [n_features]
        A numpy array containing true regression coefficient values. Returned only if `coef` is True.

    See Also
    --------
    sklearn.datasets.make_regression: non-group-sparse version
    """
    generator = check_random_state(random_state)

    total_features = n_groups * n_features_per_group
    total_informative = n_informative_groups * n_informative_per_group

    if coef:
        X, y, reg_coefs = make_regression(
            n_samples=n_samples,
            n_features=total_features,
            n_informative=total_informative,
            effective_rank=effective_rank,
            bias=0.0,
            noise=noise,
            shuffle=False,
            coef=True,
            random_state=generator,
        )
    else:
        X, y = make_regression(
            n_samples=n_samples,
            n_features=total_features,
            n_informative=total_informative,
            effective_rank=effective_rank,
            bias=0.0,
            noise=noise,
            shuffle=False,
            coef=False,
            random_state=generator,
        )

    # Evenly distribute the first `n_informative_groups * n_features_per_group`
    # features into the first `n_informative_groups` groups
    n_info_grp_features = n_informative_groups * n_features_per_group
    idx_range = np.arange(n_info_grp_features)

    idx_map_consolidated_2_grouped = (
        np.concatenate(
            [np.arange(0, n_info_grp_features, n_informative_groups)]
            * n_informative_groups
        )
        + idx_range // n_features_per_group
    )

    X = np.concatenate(
        [X[:, idx_map_consolidated_2_grouped], X[:, n_info_grp_features:]], axis=1
    )

    group_idx_map = np.concatenate(
        [np.ones(n_features_per_group, dtype=np.int32) * i for i in range(n_groups)]
    )

    if coef:
        reg_coefs = np.concatenate(
            [reg_coefs[idx_map_consolidated_2_grouped], reg_coefs[n_info_grp_features:]]
        )

    # Randomly permute samples and features
    if shuffle:
        X, y = util_shuffle(X, y, random_state=generator)

        indices = np.arange(total_features)
        generator.shuffle(indices)
        X[:, :] = X[:, indices]
        group_idx_map = group_idx_map[indices]
        if coef:
            reg_coefs = reg_coefs[indices]

    X = np.ascontiguousarray(X)
    groups = [np.where(group_idx_map == idx)[0] for idx in range(n_groups)]
    if coef:
        return X, y, groups, reg_coefs
    else:
        return X, y, groups



"""Transform feature matrices with grouped covariates."""
import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array, check_random_state
from sklearn.utils import shuffle as util_shuffle

from .utils import check_groups


def isiterable(obj):
    """Return True if obj is an iterable, False otherwise."""
    try:
        _ = iter(obj)  # noqa F841
    except TypeError:
        return False
    else:
        return True


def _check_group_names(groups, group_names):
    if group_names is not None:
        if groups is None:
            raise ValueError(
                "you provided group_names but not groups. "
                "Please provide groups also."
            )
        if isiterable(group_names) and all(isinstance(s, str) for s in group_names):
            group_names_ = np.array([set([s]) for s in group_names])
        elif isiterable(group_names) and all(isiterable(s) for s in group_names):
            group_names_ = np.array([set(s) for s in group_names])
        else:
            raise TypeError(
                "group_names must be a sequence of strings or a sequence "
                "of sequences. got {0} instead".format(group_names)
            )
        if len(group_names) != len(groups):
            raise ValueError("group_names must have the same length as groups.")
    else:
        group_names_ = None

    return group_names_


def _check_select(select, group_names_):
    missing_grp_names_msg = "if ``select`` is a string, you must provide group_names"
    if select is None:
        select_ = select
    elif isiterable(select) and all(isinstance(e, int) for e in select):
        select_ = np.array(select)
    elif isinstance(select, int):
        select_ = np.array([select])
    elif isinstance(select, str):
        if group_names_ is None:
            raise ValueError(missing_grp_names_msg)
        mask = np.array(set([select]) <= group_names_, dtype=bool)
        select_ = np.where(mask)[0]
    elif isiterable(select) and all(isinstance(e, str) for e in select):
        if group_names_ is None:
            raise ValueError(missing_grp_names_msg)

        mask = np.zeros_like(group_names_, dtype=bool)
        for label in select:
            mask = np.logical_or(mask, set([label]) <= group_names_)

        select_ = np.where(mask)[0]
    else:
        raise ValueError(
            f"``select`` must be an int, string or sequence of ints or "
            f"strings; got {select} instead."
        )

    return select_


class GroupExtractor(BaseEstimator, TransformerMixin):
    """An sklearn-compatible group extractor.

    Given a sequence of all group indices and a subsequence of desired
    group indices, this transformer returns the columns of the feature
    matrix, `X`, that are in the desired subgroups.

    Parameters
    ----------
    select : numpy.ndarray, int, or str, optional
        subsequence of desired groups to extract from feature matrix
        If int or sequence of ints, these will be treated as group indices.
        If str or sequence of str, these will be treated as labels for any
        level of the (potentially multi-indexed) group names, which must be
        specified in ``group_names``

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    group_names : sequence of str or sequences, optional
        The names of the groups of X. If this is a sequence of strings, then
        this transformer will extract groups whose names match ``select``. If
        this is a sequence of sequences, then this transformer will extract
        groups that have labels that match ``select`` at any level of their
        multi-index.

    copy_X : bool, default=False
        if ``True``, X will be copied; else, ``transform`` may return a view
    """

    def __init__(self, select=None, groups=None, group_names=None, copy_X=False):
        self.select = select
        self.groups = groups
        self.group_names = group_names
        self.copy_X = copy_X

    def transform(self, X, y=None):
        """Transform the input data, extracting the desired groups.

        Parameters
        ----------
        X : numpy.ndarray
            The feature matrix
        """
        X = check_array(
            X,
            copy=self.copy_X,
            dtype=[np.float32, np.float64, int],
            force_all_finite=False,
        )
        groups = check_groups(groups=self.groups_, X=X, allow_overlap=True)
        if self.select_ is None:
            return X

        idx = np.concatenate([groups[e] for e in self.select_])
        return X[:, idx]

    def fit(self, X=None, y=None):
        """Learn the groups and number of features from the input data."""
        X = check_array(
            X,
            copy=self.copy_X,
            dtype=[np.float32, np.float64, int],
            force_all_finite=False,
        )

        _, self.n_features_in_ = X.shape
        self.groups_ = check_groups(groups=self.groups, X=X, allow_overlap=True)
        self.group_names_ = _check_group_names(self.groups, self.group_names)
        self.select_ = _check_select(self.select, self.group_names_)

        return self

    def _more_tags(self):  # pylint: disable=no-self-use
        return {"allow_nan": True, "multilabel": True, "multioutput": True}


class GroupRemover(BaseEstimator, TransformerMixin):
    """An sklearn-compatible group remover.

    Given a sequence of all group indices and a subsequence of unwanted
    group indices, this transformer returns the columns of the feature
    matrix, `X`, that DO NOT include the unwanted subgroups.

    Parameters
    ----------
    select : numpy.ndarray, int, or str, optional
        subsequence of desired groups to remove from feature matrix
        If int or sequence of ints, these will be treated as group indices.
        If str or sequence of str, these will be treated as labels for any
        level of the (potentially multi-indexed) group names, which must be
        specified in ``group_names``

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    group_names : sequence of str or sequences, optional
        The names of the groups of X. If this is a sequence of strings, then
        this transformer will remove groups whose names match ``select``. If
        this is a sequence of sequences, then this transformer will remove
        groups that have labels that match ``select`` at any level of their
        multi-index.

    copy_X : bool, default=False
        if ``True``, X will be copied; else, ``transform`` may return a view
    """

    def __init__(self, select=None, groups=None, group_names=None, copy_X=False):
        self.select = select
        self.groups = groups
        self.group_names = group_names
        self.copy_X = copy_X

    def transform(self, X, y=None):
        """Transform the input data, removing the unwanted groups.

        Parameters
        ----------
        X : numpy.ndarray
            The feature matrix
        """
        X = check_array(
            X,
            copy=self.copy_X,
            dtype=[np.float32, np.float64, int],
            force_all_finite=False,
        )
        groups = check_groups(groups=self.groups_, X=X, allow_overlap=True)
        if self.select_ is None:
            return X

        idx = np.concatenate(
            [grp for idx, grp in enumerate(groups) if idx not in self.select_]
        )
        return X[:, idx]

    def fit(self, X=None, y=None):
        """Learn the groups and number of features from the input data."""
        X = check_array(
            X,
            copy=self.copy_X,
            dtype=[np.float32, np.float64, int],
            force_all_finite=False,
        )

        _, self.n_features_in_ = X.shape
        self.groups_ = check_groups(groups=self.groups, X=X, allow_overlap=True)
        self.group_names_ = _check_group_names(self.groups, self.group_names)
        self.select_ = _check_select(self.select, self.group_names_)

        return self

    def _more_tags(self):  # pylint: disable=no-self-use
        return {"allow_nan": True, "multilabel": True, "multioutput": True}


class GroupShuffler(BaseEstimator, TransformerMixin):
    """Shuffle some groups of a feature matrix, leaving others as is.

    Given a sequence of all group indices and a subsequence of
    group indices, this transformer returns the feature
    matrix, `X`, with the subset of groups shuffled.

    Parameters
    ----------
    select : numpy.ndarray, int, or str, optional
        subsequence of desired groups to shuffle in the feature matrix
        If int or sequence of ints, these will be treated as group indices.
        If str or sequence of str, these will be treated as labels for any
        level of the (potentially multi-indexed) group names, which must be
        specified in ``group_names``

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    group_names : sequence of str or sequences, optional
        The names of the groups of X. If this is a sequence of strings, then
        this transformer will shuffle groups whose names match ``select``. If
        this is a sequence of sequences, then this transformer will shuffle
        groups that have labels that match ``select`` at any level of their
        multi-index.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    """

    def __init__(self, select=None, groups=None, group_names=None, random_state=None):
        self.select = select
        self.groups = groups
        self.group_names = group_names
        self.random_state = random_state

    def transform(self, X, y=None):
        """Transform the input data, removing the unwanted groups.

        Parameters
        ----------
        X : numpy.ndarray
            The feature matrix
        """
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=False
        )
        groups = check_groups(groups=self.groups_, X=X, allow_overlap=True)
        if self.select_ is None:
            return X

        generator = check_random_state(self.random_state)
        idx = np.concatenate([groups[e] for e in self.select_])
        shuffle_view = X[:, idx]
        shuffle_view = util_shuffle(shuffle_view, random_state=generator)
        X[:, idx] = shuffle_view
        return X

    def fit(self, X=None, y=None):
        """Learn the groups and number of features from the input data."""
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=False
        )

        _, self.n_features_in_ = X.shape
        self.groups_ = check_groups(groups=self.groups, X=X, allow_overlap=True)
        self.group_names_ = _check_group_names(self.groups, self.group_names)
        self.select_ = _check_select(self.select, self.group_names_)

        return self

    def _more_tags(self):  # pylint: disable=no-self-use
        return {"allow_nan": True, "multilabel": True, "multioutput": True}



"""Perform group-wise functional PCA of feature matrices with grouped covariates."""
import inspect
import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA
from sklearn.feature_selection import SelectKBest, f_regression, f_classif
from sklearn.utils.multiclass import type_of_target
from sklearn.utils import check_array, check_scalar

try:
    from skfda import FDataGrid
    from skfda.representation.basis import Basis
    from skfda.representation.basis import BSpline, Monomial, Constant, Fourier
    from skfda.preprocessing.dim_reduction.projection import FPCA

    HAS_SKFDA = True
except ImportError:  # pragma: no cover
    HAS_SKFDA = False

from .utils import check_groups


class GroupPCA(BaseEstimator, TransformerMixin):
    """An sklearn-compatible grouped PCA transformer.

    Given a feature matrix `X` and grouping information,
    this transformer returns the groupwise PCA matrix.

    Parameters
    ----------
    n_components : int, float or 'mle', default=None
        Number of components to keep for each group.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features_in_group)

        If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
        MLE is used to guess the dimension. Use of ``n_components == 'mle'``
        will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.

        If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
        number of components such that the amount of variance that needs to be
        explained is greater than the percentage specified by n_components.

        If ``svd_solver == 'arpack'``, the number of components must be
        strictly less than the minimum of n_features_in_group and n_samples.

        Hence, the None case results in::

            n_components == min(n_samples, n_features_in_group) - 1

    whiten : bool, default=False
        When True (False by default) the `components_` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometimes
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    svd_solver : {'auto', 'full', 'arpack', 'randomized'}, default='auto'
        If auto :
            The solver is selected by a default policy based on `X.shape` and
            `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        If full :
            run exact full SVD calling the standard LAPACK solver via
            `scipy.linalg.svd` and select the components by postprocessing
        If arpack :
            run SVD truncated to n_components calling ARPACK solver via
            `scipy.sparse.linalg.svds`. It requires strictly
            0 < n_components < min(X.shape)
        If randomized :
            run randomized SVD by the method of Halko et al.

    tol : float, default=0.0
        Tolerance for singular values computed by svd_solver == 'arpack'.
        Must be of range [0.0, infinity).

    iterated_power : int or 'auto', default='auto'
        Number of iterations for the power method computed by
        svd_solver == 'randomized'.
        Must be of range [0, infinity).

    random_state : int, RandomState instance or None, default=None
        Used when the 'arpack' or 'randomized' solvers are used. Pass an int
        for reproducible results across multiple function calls.

    groups : numpy.ndarray or int, optional
        all group indices for feature matrix

    exclude_groups : sequence of int or int, optional
        a sequence of group indices (or a single group index) to exclude from
        PCA transformatiton. These groups will be left untouched and
        transferred to the output matrix.

    Attributes
    ----------
    components_ : list of ndarray of shape (n_components, n_features) for each group
        List of each group's principal axes in feature space, representing
        the directions of maximum variance in the data. The components are
        sorted by ``explained_variance_``.

    explained_variance_ : list of ndarray of shape (n_components,) for each group
        List of each group's amount of variance explained by each of the
        selected components. Equal to n_components largest eigenvalues of the
        covariance matrix of X.

    explained_variance_ratio_ : list of ndarray of shape (n_components,) for each group
        Percentage of variance explained by each of the selected components.
        If ``n_components`` is not set then all components are stored and the
        sum of the ratios is equal to 1.0.

    singular_values_ : list of ndarray of shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    mean_ : list of ndarray of shape (n_features,)
        Per-feature empirical mean, estimated from the training set.
        Equal to `X.mean(axis=0)`.

    n_components_ : list of int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features_per_group and n_samples
        if n_components is None.

    noise_variance_ : float
        The estimated noise covariance following the Probabilistic PCA model
        from Tipping and Bishop 1999. See "Pattern Recognition and
        Machine Learning" by C. Bishop, 12.2.1 p. 574 or
        http://www.miketipping.com/papers/met-mppca.pdf. It is required to
        compute the estimated data covariance and score samples.

        Equal to the average of (min(n_features, n_samples) - n_components)
        smallest eigenvalues of the covariance matrix of X.

    groups_in_ : list of numpy.ndarray
        explicit group indices created from the potentially more implicit
        ``groups`` input parameter

    groups_out_ : list of numpy.ndarray
        group indices for the transformed features

    n_features_in_ : int
        number of input features

    n_features_out_ : int
        number of transformed features

    n_samples_ : int
        Number of samples in the training data.

    """

    def __init__(
        self,
        n_components=None,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
        groups=None,
        exclude_groups=None,
    ):
        self.n_components = n_components
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state
        self.groups = groups
        self.exclude_groups = exclude_groups

    def _check_exclude_groups(self):
        if self.exclude_groups is None:
            return []
        elif isinstance(self.exclude_groups, int):
            return [self.exclude_groups]
        elif all(isinstance(e, int) for e in self.exclude_groups):
            return list(self.exclude_groups)
        else:
            raise TypeError(
                f"exclude_groups must be and int or sequence of ints. "
                f"Got {self.exclude_groups} instead."
            )

    def fit(self, X, y=None):
        """Fit the GroupPCA model with X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=True
        )

        self.n_samples_, self.n_features_in_ = X.shape
        self.groups_in_ = check_groups(groups=self.groups, X=X, allow_overlap=False)

        exclude_grp_idx = self._check_exclude_groups()

        self.pca_models_ = [
            PCA(
                n_components=self.n_components,
                copy=True,
                whiten=self.whiten,
                svd_solver=self.svd_solver,
                tol=self.tol,
                iterated_power=self.iterated_power,
                random_state=self.random_state,
            )
            for grp in self.groups_in_
        ]
        self.components_ = [None for grp in self.groups_in_]
        self.explained_variance_ = [None for grp in self.groups_in_]
        self.explained_variance_ratio_ = [None for grp in self.groups_in_]
        self.singular_values_ = [None for grp in self.groups_in_]
        self.mean_ = [None for grp in self.groups_in_]
        self.n_components_ = [None for grp in self.groups_in_]
        self.noise_variance_ = [None for grp in self.groups_in_]
        self.groups_out_ = []

        feature_start_idx = 0
        for idx, grp in enumerate(self.groups_in_):
            if idx not in exclude_grp_idx:
                group_X = np.copy(X[:, grp])
                self.pca_models_[idx].fit(group_X)
                self.components_[idx] = self.pca_models_[idx].components_
                self.explained_variance_[idx] = self.pca_models_[
                    idx
                ].explained_variance_
                self.explained_variance_ratio_[idx] = self.pca_models_[
                    idx
                ].explained_variance_ratio_
                self.singular_values_[idx] = self.pca_models_[idx].singular_values_
                self.mean_[idx] = self.pca_models_[idx].mean_
                self.n_components_[idx] = self.pca_models_[idx].n_components_
                self.noise_variance_[idx] = self.pca_models_[idx].noise_variance_
                self.groups_out_.append(
                    np.arange(
                        feature_start_idx,
                        feature_start_idx + self.pca_models_[idx].n_components_,
                    )
                )
                feature_start_idx += self.pca_models_[idx].n_components_
            else:
                self.pca_models_[idx] = None
                # self.components_ is already set to None above
                self.groups_out_.append(
                    np.arange(feature_start_idx, feature_start_idx + len(grp))
                )
                feature_start_idx += len(grp)

        self.n_features_out_ = np.sum([len(grp) for grp in self.groups_out_])

        return self

    def transform(self, X, y=None):
        """Apply dimensionality reduction to X.

        X is projected on the principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=True
        )
        groups = check_groups(groups=self.groups_in_, X=X, allow_overlap=False)

        X_out = []

        for idx, grp in enumerate(groups):
            if self.pca_models_[idx] is not None:
                X_out.append(self.pca_models_[idx].transform(X[:, grp]))
            else:
                X_out.append(X[:, grp])

        return np.hstack(X_out)

    def inverse_transform(self, X):
        """Transform data back to its original space.

        In other words, return an input X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples is the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original array-like, shape (n_samples, n_features)

        Notes
        -----
        If whitening is enabled, inverse_transform will compute the
        exact inverse operation, which includes reversing whitening.
        """
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=True
        )
        groups = check_groups(groups=self.groups_out_, X=X, allow_overlap=False)

        X_out = []

        for idx, grp in enumerate(groups):
            if self.pca_models_[idx] is not None:
                X_out.append(self.pca_models_[idx].inverse_transform(X[:, grp]))
            else:
                X_out.append(X[:, grp])

        return np.hstack(X_out)

    def _more_tags(self):  # pylint: disable=no-self-use
        return {
            "allow_nan": False,
            "multilabel": True,
            "multioutput": True,
            "_skip_test": True,
        }


def _get_score_func(y):
    target_type = type_of_target(y)
    if target_type == "continuous":
        return f_regression
    elif target_type in ["binary", "multiclass"]:
        return f_classif
    else:
        raise NotImplementedError(
            "The type of target is unsupported. It must be continuous, "
            "binary, or multiclass. Got {0} instead.".format(target_type)
        )


class GroupSDRMixin:
    """Mixin class for grouped supervised dimensionality reduction (SDR)."""

    def _compute_screening_mask(self, X, y):
        y = check_array(
            y,
            copy=True,
            dtype=[np.float32, np.float64, int],
            force_all_finite=True,
            ensure_2d=False,
        )
        if y.ndim != 1:
            raise NotImplementedError("y must be one-dimensional.")

        self.univariate_score_func_ = _get_score_func(y)
        self.groups_in_ = check_groups(groups=self.groups, X=X, allow_overlap=False)
        exclude_grp_idx = self._check_exclude_groups()

        check_scalar(
            self.theta,
            name="theta",
            target_type=(float, np.float64, np.float32, int),
            min_val=0.0,
            max_val=1.0,
        )
        if self.theta == 0.0:
            self.screening_mask_ = np.ones(X.shape[1])
            return

        screener = SelectKBest(self.univariate_score_func_)
        screening_masks = []
        for idx, grp in enumerate(self.groups_in_):
            if idx not in exclude_grp_idx:
                screener.fit(X[:, grp], y)
                univariate_coefs = np.abs(screener.scores_)

                # Create a mask where the univariate coefs are above the theta threshold
                if self.absolute_threshold:
                    mask = univariate_coefs >= self.theta
                else:
                    mask = univariate_coefs >= self.theta * np.max(univariate_coefs)

                # Find where mask transitions between 0 and 1
                m = mask.astype(np.float64)
                lower = m < 0.5
                higher = m > 0.5
                transition_indices = np.sort(
                    np.concatenate(
                        [
                            np.where(lower[:-1] & higher[1:])[0],
                            np.where(higher[:-1] & lower[1:])[0],
                        ]
                    )
                )

                # Create a "soft" mask where values above the theta threshold
                # are one and they gradually fall to zero outside
                if len(transition_indices) and self.smooth_screening:
                    dist_from_transition = np.min(
                        np.abs(np.arange(len(m)) - transition_indices[:, np.newaxis]),
                        axis=0,
                    )
                    dist_from_transition[mask] = 0
                    soft_mask = np.exp(-np.square(dist_from_transition) / 4)
                else:
                    soft_mask = np.copy(m)

                screening_masks.append(soft_mask)
            else:
                screening_masks.append(np.ones(len(grp)))

        self.screening_mask_ = np.concatenate(screening_masks)


class SupervisedGroupPCA(GroupPCA, GroupSDRMixin):
    """A grouped supervised PCA transformer.

    Given a feature matrix `X`, grouping information, and a supervision
    target this transformer returns the groupwise supervised PCA
    matrix.

    Parameters
    ----------
    n_components : int, float or 'mle', default=None
        Number of components to keep for each group.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features_in_group)

        If ``n_components == 'mle'`` and ``svd_solver == 'full'``, Minka's
        MLE is used to guess the dimension. Use of ``n_components == 'mle'``
        will interpret ``svd_solver == 'auto'`` as ``svd_solver == 'full'``.

        If ``0 < n_components < 1`` and ``svd_solver == 'full'``, select the
        number of components such that the amount of variance that needs to be
        explained is greater than the percentage specified by n_components.

        If ``svd_solver == 'arpack'``, the number of components must be
        strictly less than the minimum of n_features_in_group and n_samples.

        Hence, the None case results in::

            n_components == min(n_samples, n_features_in_group) - 1

    theta : float, default=0.0
        the threshold for each features univariate regression coefficient. If
        the absolute value of an individual feature's univariate regression
        coefficient is below this threshold, its value for all samples is set
        to the mean feature value. Unless ``absolute_threshold`` is True,
        theta is scaled by the max absolute value of the regression
        coefficients in each group.

    absolute_threshold : boolean, default=False
        If True, treat ``theta`` as an absolute threshold. i.e. do not scale
        by the max absolute regression coefficient in each group.

    smooth_screening : bool, default=False
        If True, values below the `theta` threshold fall smoothly to the mean
        feature value. If False, they are set abruptly to the mean.

    whiten : bool, default=False
        When True (False by default) the `components_` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometimes
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    svd_solver : {'auto', 'full', 'arpack', 'randomized'}, default='auto'
        If auto :
            The solver is selected by a default policy based on `X.shape` and
            `n_components`: if the input data is larger than 500x500 and the
            number of components to extract is lower than 80% of the smallest
            dimension of the data, then the more efficient 'randomized'
            method is enabled. Otherwise the exact full SVD is computed and
            optionally truncated afterwards.
        If full :
            run exact full SVD calling the standard LAPACK solver via
            `scipy.linalg.svd` and select the components by postprocessing
        If arpack :
            run SVD truncated to n_components calling ARPACK solver via
            `scipy.sparse.linalg.svds`. It requires strictly
            0 < n_components < min(X.shape)
        If randomized :
            run randomized SVD by the method of Halko et al.

    tol : float, default=0.0
        Tolerance for singular values computed by svd_solver == 'arpack'.
        Must be of range [0.0, infinity).

    iterated_power : int or 'auto', default='auto'
        Number of iterations for the power method computed by
        svd_solver == 'randomized'.
        Must be of range [0, infinity).

    random_state : int, RandomState instance or None, default=None
        Used when the 'arpack' or 'randomized' solvers are used. Pass an int
        for reproducible results across multiple function calls.

    groups : numpy.ndarray or int, optional
        all group indices for feature matrix

    exclude_groups : sequence of int or int, optional
        a sequence of group indices (or a single group index) to exclude from
        PCA transformatiton. These groups will be left untouched and
        transferred to the output matrix.

    Attributes
    ----------
    components_ : list of ndarray of shape (n_components, n_features) for each group
        List of each group's principal axes in feature space, representing
        the directions of maximum variance in the data. The components are
        sorted by ``explained_variance_``.

    explained_variance_ : list of ndarray of shape (n_components,) for each group
        List of each group's amount of variance explained by each of the
        selected components. Equal to n_components largest eigenvalues of the
        covariance matrix of X.

    explained_variance_ratio_ : list of ndarray of shape (n_components,) for each group
        Percentage of variance explained by each of the selected components.
        If ``n_components`` is not set then all components are stored and the
        sum of the ratios is equal to 1.0.

    singular_values_ : list of ndarray of shape (n_components,)
        The singular values corresponding to each of the selected components.
        The singular values are equal to the 2-norms of the ``n_components``
        variables in the lower-dimensional space.

    mean_ : list of ndarray of shape (n_features,)
        Per-feature empirical mean, estimated from the training set.
        Equal to `X.mean(axis=0)`.

    n_components_ : list of int
        The estimated number of components. When n_components is set
        to 'mle' or a number between 0 and 1 (with svd_solver == 'full') this
        number is estimated from input data. Otherwise it equals the parameter
        n_components, or the lesser value of n_features_per_group and n_samples
        if n_components is None.

    noise_variance_ : float
        The estimated noise covariance following the Probabilistic PCA model
        from Tipping and Bishop 1999. See "Pattern Recognition and
        Machine Learning" by C. Bishop, 12.2.1 p. 574 or
        http://www.miketipping.com/papers/met-mppca.pdf. It is required to
        compute the estimated data covariance and score samples.

        Equal to the average of (min(n_features, n_samples) - n_components)
        smallest eigenvalues of the covariance matrix of X.

    groups_in_ : list of numpy.ndarray
        explicit group indices created from the potentially more implicit
        ``groups`` input parameter

    groups_out_ : list of numpy.ndarray
        group indices for the transformed features

    n_features_in_ : int
        number of input features

    n_features_out_ : int
        number of transformed features

    n_samples_ : int
        Number of samples in the training data.

    """

    def __init__(
        self,
        n_components=None,
        theta=0.0,
        absolute_threshold=False,
        smooth_screening=False,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
        groups=None,
        exclude_groups=None,
    ):
        self.n_components = n_components
        self.theta = theta
        self.absolute_threshold = absolute_threshold
        self.smooth_screening = smooth_screening
        self.whiten = whiten
        self.svd_solver = svd_solver
        self.tol = tol
        self.iterated_power = iterated_power
        self.random_state = random_state
        self.groups = groups
        self.exclude_groups = exclude_groups

    def fit(self, X, y):
        """Fit the GroupPCA model with X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        y : Ignored

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._compute_screening_mask(X, y)
        return super().fit(X=X * self.screening_mask_)

    def transform(self, X, y=None):
        """Apply dimensionality reduction to X.

        X is projected on the principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        return super().transform(X * self.screening_mask_)

    def _more_tags(self):  # pylint: disable=no-self-use
        return {
            "allow_nan": False,
            "multilabel": True,
            "multioutput": True,
            "_skip_test": True,
            "requires_y": True,
        }


def _check_skfda():
    if not HAS_SKFDA:
        raise ImportError(
            "To use functional data analysis in groupyr, you will need to have "
            "scikit-fda installed. You can do this by installing groupyr with "
            "`pip install groupyr[fda]`, or by separately installing scikit-fda "
            "with `pip install scikit-fda`."
        )


def _has_nbasis_kwarg(_basis):
    try:
        # Not all bases have an ``n_basis`` kwarg
        # If we just tried to instantiate the basis, then
        # we'd catch all TypeErrors raised by ``basis``.
        # Instead, we use ``inspect``
        inspect.signature(_basis).bind(n_basis=7)
        return True
    except TypeError:
        return False


def _check_basis(basis, arg_name, n_basis, domain_range=None):
    _check_skfda()
    allowed_bases = {
        "bspline": BSpline,
        "monomial": Monomial,
        "constant": Constant,
        "fourier": Fourier,
    }

    err_msg = (
        f"{arg_name} must be one of {list(allowed_bases.keys())} or be an "
        f"instance of a basis class from `skfda.representation.basis`."
    )

    if inspect.isclass(basis):
        if not issubclass(basis, Basis):
            raise ValueError(err_msg)

        if _has_nbasis_kwarg(basis):
            return basis(n_basis=n_basis, domain_range=domain_range)
        else:
            return basis(domain_range=domain_range)
    elif basis is not None:
        if basis.lower() not in allowed_bases.keys():
            raise ValueError(err_msg)

        if _has_nbasis_kwarg(allowed_bases[basis.lower()]):
            return allowed_bases[basis.lower()](
                n_basis=n_basis, domain_range=domain_range
            )
        else:
            return allowed_bases[basis.lower()](domain_range=domain_range)

    return None


def _get_group_fd(X, group_mask, basis, grid_points=None):
    _check_skfda()
    group_X = np.copy(X[:, group_mask])
    if grid_points is None:
        group_grid_points = np.arange(0, group_X.shape[1])
    else:
        group_grid_points = np.copy(grid_points[group_mask])
    fd = FDataGrid(group_X, group_grid_points)
    if basis is not None:
        fd = fd.to_basis(basis)
    return fd


class GroupFPCA(BaseEstimator, TransformerMixin):
    """An sklearn-compatible grouped functional PCA transformer.

    Given a feature matrix `X` and grouping information,
    this transformer returns the groupwise functional PCA
    matrix.

    Parameters
    ----------
    n_components : int, default=3
        number of principal components to obtain from functional
        principal component analysis.

    centering : bool, default=True
        if True then calculate the mean of the functional data object and
        center the data first.

    basis : str, instance of `skfda.representation.basis` class, optional
        the basis in which to represent each group's function before fPCA.

    n_basis : int, default=4
        the number of functions in the basis. Only used if ``basis`` is not
        None.

    basis_domain_range : tuple, optional
        a tuple of length 2 containing the initial and end values of the
        interval over which the basis can be evaluated.

    groups : numpy.ndarray or int, optional
        all group indices for feature matrix

    exclude_groups : sequence of int or int, optional
        a sequence of group indices (or a single group index) to exclude from
        fPCA transformatiton. These groups will be left untouched and
        transferred to the output matrix.

    Attributes
    ----------
    components_ : list of skfda.FData
        a list where each element contains a group's principal components
        in a basis representation.

    fitted_grid_points_ : numpy.ndarray or None
        The dicretization points for the input ``X`` matrix
        provided during ``fit``.

    fpca_models_ : list of fitted skfda.FPCA models
        a list where each element contains a group's fitted FPCA model

    groups_in_ : list of numpy.ndarray
        explicit group indices created from the potentially more implicit
        ``groups`` input parameter

    groups_out_ : list of numpy.ndarray
        group indices for the transformed features

    n_features_in_ : int
        number of input features

    n_features_out_ : int
        number of transformed features

    """

    def __init__(
        self,
        n_components=3,
        centering=True,
        basis=None,
        n_basis=4,
        basis_domain_range=None,
        groups=None,
        exclude_groups=None,
    ):
        self.n_components = n_components
        self.centering = centering
        self.basis = basis
        self.n_basis = n_basis
        self.basis_domain_range = basis_domain_range
        self.groups = groups
        self.exclude_groups = exclude_groups

    def _check_exclude_groups(self):
        if self.exclude_groups is None:
            return []
        elif isinstance(self.exclude_groups, int):
            return [self.exclude_groups]
        elif all(isinstance(e, int) for e in self.exclude_groups):
            return list(self.exclude_groups)
        else:
            raise TypeError(
                f"exclude_groups must be and int or sequence of ints. "
                f"Got {self.exclude_groups} instead."
            )

    def fit(self, X, y=None, grid_points=None):
        """Fit the fPCA transformer.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            training vectors, where n_samples is the number of samples and
            n_features is the number of predictors.

        grid_points : numpy.ndarray, optional
            The points of dicretisation for the input ``X`` matrix. If None, this will use the
            grid_points supplied during the fit method.
        """
        _check_skfda()
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=True
        )
        basis = _check_basis(self.basis, "basis", self.n_basis, self.basis_domain_range)

        _, self.n_features_in_ = X.shape
        self.groups_in_ = check_groups(groups=self.groups, X=X, allow_overlap=False)

        if grid_points is None:
            self.fitted_grid_points_ = grid_points
        else:
            _ = check_groups(
                groups=self.groups_in_,
                X=grid_points,
                allow_overlap=False,
                kwarg_name="grid_points",
            )
            self.fitted_grid_points_ = np.array(grid_points)

        exclude_grp_idx = self._check_exclude_groups()

        self.fpca_models_ = [
            FPCA(n_components=self.n_components, centering=self.centering)
            for grp in self.groups_in_
        ]
        self.components_ = [None for grp in self.groups_in_]
        self.groups_out_ = []

        feature_start_idx = 0
        for idx, grp in enumerate(self.groups_in_):
            if idx not in exclude_grp_idx:
                fd = _get_group_fd(X, grp, basis, self.fitted_grid_points_)
                self.fpca_models_[idx].fit(fd)
                self.components_[idx] = self.fpca_models_[idx].components_
                self.groups_out_.append(
                    np.arange(feature_start_idx, feature_start_idx + self.n_components)
                )
                feature_start_idx += self.n_components
            else:
                self.fpca_models_[idx] = None
                # self.components_ is already set to None above
                self.groups_out_.append(
                    np.arange(feature_start_idx, feature_start_idx + len(grp))
                )
                feature_start_idx += len(grp)

        self.n_features_out_ = np.sum([len(grp) for grp in self.groups_out_])

        return self

    def transform(self, X, y=None, grid_points=None):
        """Transform the input data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            samples to transform.

        grid_points : numpy.ndarray, optional
            The points of dicretisation for the input ``X`` matrix. If None, this will use the
            grid_points supplied during the fit method.
        """
        X = check_array(
            X, copy=True, dtype=[np.float32, np.float64, int], force_all_finite=True
        )
        basis = _check_basis(self.basis, "basis", self.n_basis, self.basis_domain_range)
        groups = check_groups(groups=self.groups_in_, X=X, allow_overlap=False)

        X_out = []

        if grid_points is None:
            transform_grid_points = self.fitted_grid_points_
        else:
            _ = check_groups(
                groups=self.groups_in_,
                X=grid_points,
                allow_overlap=False,
                kwarg_name="grid_points",
            )
            transform_grid_points = np.array(grid_points)

        for idx, grp in enumerate(groups):
            if self.fpca_models_[idx] is not None:
                fd = _get_group_fd(X, grp, basis, transform_grid_points)
                X_out.append(self.fpca_models_[idx].transform(fd))
            else:
                X_out.append(X[:, grp])

        return np.hstack(X_out)

    def _more_tags(self):  # pylint: disable=no-self-use
        return {
            "allow_nan": False,
            "multilabel": True,
            "multioutput": True,
            "_skip_test": True,
        }


class SupervisedGroupFPCA(GroupFPCA, GroupSDRMixin):
    """A grouped supervised functional PCA transformer.

    Given a feature matrix `X`, grouping information, and a supervision
    target this transformer returns the groupwise supervised functional PCA
    matrix.

    Parameters
    ----------
    n_components : int, default=3
        number of principal components to obtain from functional
        principal component analysis.

    theta : float, default=0.0
        the threshold for each features univariate regression coefficient. If
        the absolute value of an individual feature's univariate regression
        coefficient is below this threshold, its value for all samples is set
        to the mean feature value. Unless ``absolute_threshold`` is True,
        theta is scaled by the max absolute value of the regression
        coefficients in each group.

    absolute_threshold : boolean, default=False
        If True, treat ``theta`` as an absolute threshold. i.e. do not scale
        by the max absolute regression coefficient in each group.

    smooth_screening : bool, default=True
        If True, values below the `theta` threshold fall smoothly to the mean
        feature value. If False, they are set abruptly to the mean.

    centering : bool, default=True
        if True then calculate the mean of the functional data object and
        center the data first.

    basis : str, instance of `skfda.representation.basis` class, optional
        the basis in which to represent each group's function before fPCA.

    n_basis : int, default=4
        the number of functions in the basis. Only used if ``basis`` is not
        None.

    basis_domain_range : tuple, optional
        a tuple of length 2 containing the initial and end values of the
        interval over which the basis can be evaluated.

    groups : numpy.ndarray or int, optional
        all group indices for feature matrix

    exclude_groups : sequence of int or int, optional
        a sequence of group indices (or a single group index) to exclude from
        fPCA transformatiton. These groups will be left untouched and
        transferred to the output matrix.

    Attributes
    ----------
    components_ : list of skfda.FData
        a list where each element contains a group's principal components
        in a basis representation.

    fitted_grid_points_ : numpy.ndarray or None
        The dicretization points for the input ``X`` matrix
        provided during ``fit``.

    fpca_models_ : list of fitted skfda.FPCA models
        a list where each element contains a group's fitted FPCA model

    groups_in_ : list of numpy.ndarray
        explicit group indices created from the potentially more implicit
        ``groups`` input parameter

    groups_out_ : list of numpy.ndarray
        group indices for the transformed features

    n_features_in_ : int
        number of input features

    n_features_out_ : int
        number of transformed features

    screening_mask_ : np.ndarray

    """

    def __init__(
        self,
        n_components=3,
        theta=0.0,
        absolute_threshold=False,
        smooth_screening=True,
        centering=True,
        basis=None,
        n_basis=4,
        basis_domain_range=None,
        groups=None,
        exclude_groups=None,
    ):
        self.theta = theta
        self.absolute_threshold = absolute_threshold
        self.smooth_screening = smooth_screening
        super().__init__(
            n_components=n_components,
            centering=centering,
            basis=basis,
            n_basis=n_basis,
            basis_domain_range=basis_domain_range,
            groups=groups,
            exclude_groups=exclude_groups,
        )

    def fit(self, X, y, grid_points=None):
        """Fit the fPCA transformer.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            training vectors, where n_samples is the number of samples and
            n_features is the number of predictors.

        y : array-like of shape (n_samples,)
            target vectors, where n_samples is the number of samples

        grid_points : numpy.ndarray
            The points of dicretisation for the input ``X`` matrix. If None, this will use the
            grid_points supplied during the fit method.
        """
        self._compute_screening_mask(X, y)
        return super().fit(X=X * self.screening_mask_, grid_points=grid_points)

    def transform(self, X, y=None, grid_points=None):
        """Transform the input data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            samples to transform.

        grid_points : numpy.ndarray, optional
            The points of dicretisation for the input ``X`` matrix. If None, this will use the
            grid_points supplied during the fit method.
        """
        return super().transform(X * self.screening_mask_, grid_points=grid_points)

    def _more_tags(self):  # pylint: disable=no-self-use
        return {
            "allow_nan": False,
            "multilabel": True,
            "multioutput": True,
            "_skip_test": True,
            "requires_y": True,
        }



"""Create regression estimators based on the sparse group lasso."""
import contextlib
import logging
import numpy as np
import warnings

from functools import partial
from joblib import delayed, effective_n_jobs, Parallel
from scipy import sparse
from scipy.optimize import root_scalar
from skopt import BayesSearchCV
from tqdm.auto import tqdm

from sklearn.base import RegressorMixin, TransformerMixin
from sklearn.linear_model._base import LinearModel, _preprocess_data
from sklearn.linear_model._coordinate_descent import _alpha_grid as _lasso_alpha_grid
from sklearn.metrics import get_scorer
from sklearn.model_selection import check_cv
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.fixes import _joblib_parallel_args
from sklearn.utils.validation import check_array, check_is_fitted, column_or_1d

from ._base import SGLBaseEstimator
from ._prox import _soft_threshold
from .utils import check_groups

__all__ = ["SGL", "sgl_path", "SGLCV"]
logger = logging.getLogger(__name__)


class SGL(SGLBaseEstimator, RegressorMixin, LinearModel):
    """An sklearn compatible sparse group lasso regressor.

    This solves the sparse group lasso [1]_ problem for a feature matrix
    partitioned into groups using the proximal gradient descent (PGD)
    algorithm.

    Parameters
    ----------
    l1_ratio : float, default=1.0
        Hyper-parameter : Combination between group lasso and lasso. l1_ratio=0
        gives the group lasso and l1_ratio=1 gives the lasso.

    alpha : float, default=1.0
        Hyper-parameter : overall regularization strength.

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group. We set groups in ``__init__`` so
        that it can be reused in model selection and CV routines.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    fit_intercept : bool, default=True
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the linear predictor (X @ coef + intercept).

    max_iter : int, default=1000
        Maximum number of iterations for PGD solver.

    tol : float, default=1e-7
        Stopping criterion. Convergence tolerance for the ``copt`` proximal gradient solver

    warm_start : bool, default=False
        If set to ``True``, reuse the solution of the previous call to ``fit``
        as initialization for ``coef_`` and ``intercept_``.

    verbose : int, default=0
        Verbosity flag for PGD solver. Any positive integer will produce
        verbose output

    suppress_solver_warnings : bool, default=True
        If True, suppress convergence warnings from PGD solver.
        This is useful for hyperparameter tuning when some combinations
        of hyperparameters may not converge.

    Attributes
    ----------
    coef_ : array of shape (n_features,)
        Estimated coefficients for the linear predictor (`X @ coef_ +
        intercept_`).

    intercept_ : float
        Intercept (a.k.a. bias) added to linear predictor.

    n_iter_ : int
        Actual number of iterations used in the solver.

    References
    ----------
    .. [1]  Noah Simon, Jerome Friedman, Trevor Hastie & Robert Tibshirani,
        "A Sparse-Group Lasso," Journal of Computational and Graphical
        Statistics, vol. 22:2, pp. 231-245, 2012
        DOI: 10.1080/10618600.2012.681250

    """

    def fit(self, X, y, loss="squared_loss"):
        """Fit a linear model using the sparse group lasso.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.

        y : array-like, shape (n_samples,)
            The target values (class labels in classification, real numbers in
            regression).

        loss : ["squared_loss", "huber"]
            The type of loss function to use in the PGD solver.

        Returns
        -------
        self : object
            Returns self.
        """
        return super().fit(X=X, y=y, loss=loss)

    def predict(self, X):
        """Predict targets for test vectors in ``X``.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            Returns an array of ones.
        """
        X = check_array(X, accept_sparse=True)
        check_is_fitted(self, "is_fitted_")

        return safe_sparse_dot(X, self.coef_.T, dense_output=True) + self.intercept_

    def _more_tags(self):  # pylint: disable=no-self-use
        return {"requires_y": True}


def _alpha_grid(
    X,
    y,
    Xy=None,
    groups=None,
    scale_l2_by="group_length",
    l1_ratio=1.0,
    fit_intercept=True,
    eps=1e-3,
    n_alphas=100,
    normalize=False,
    copy_X=True,
    model=SGL,
):
    """Compute the grid of alpha values for elastic net parameter search.

    Parameters
    ----------
    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        Training data. Pass directly as Fortran-contiguous data to avoid
        unnecessary memory duplication

    y : ndarray of shape (n_samples,)
        Target values

    Xy : array-like of shape (n_features,), default=None
        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
        prevent train/test leakage by ensuring the ``Xy`` is precomputed
        using only training data.

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    l1_ratio : float, default=1.0
        The elastic net mixing parameter, with ``0 < l1_ratio <= 1``.
        For ``l1_ratio = 0`` the penalty is an L2 penalty. (currently not
        supported) ``For l1_ratio = 1`` it is an L1 penalty. For
        ``0 < l1_ratio <1``, the penalty is a combination of L1 and L2.

    eps : float, default=1e-3
        Length of the path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``

    n_alphas : int, default=100
        Number of alphas along the regularization path

    fit_intercept : bool, default=True
        Whether to fit an intercept or not

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : bool, default=True
        If ``True``, X will be copied; else, it may be overwritten.

    model : class, default=SGL
        The estimator class that will be used to confirm that alpha_max sets
        all coef values to zero. The default value of ``model=SGL`` is
        appropriate for regression while ``model=LogisticSGL`` is appropriate
        for classification.
    """
    if l1_ratio == 1.0:
        return _lasso_alpha_grid(
            X=X,
            y=y,
            Xy=Xy,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            eps=eps,
            n_alphas=n_alphas,
            normalize=normalize,
            copy_X=copy_X,
        )

    n_samples = len(y)
    if Xy is None:
        X = check_array(X, accept_sparse=False, copy=(copy_X and fit_intercept))
        X, y, _, _, _ = _preprocess_data(X, y, fit_intercept, normalize, copy=False)
        Xy = safe_sparse_dot(X.T, y, dense_output=True)

    if Xy.ndim == 1:
        Xy = Xy[:, np.newaxis]

    groups = check_groups(groups, X, allow_overlap=False, fit_intercept=False)

    if scale_l2_by not in ["group_length", None]:
        raise ValueError(
            "scale_l2_by must be 'group_length' or None; " "got {0}".format(scale_l2_by)
        )

    # When l1_ratio < 1 (i.e. not the lasso), then for each group, the
    # smallest alpha for which coef_ = 0 minimizes the objective will be
    # achieved when
    #
    # || S(Xy / n_samples, l1_ratio * alpha) ||_2 == sqrt(p_l) * (1 - l1_ratio) * alpha
    #
    # where S() is the element-wise soft-thresholding operator and p_l is
    # the group size (or 1 if ``scale_l2_by is None``)
    def beta_zero_root(alpha, group):
        soft = _soft_threshold(Xy[group] / n_samples, l1_ratio * alpha)
        scale = np.sqrt(group.size) if scale_l2_by == "group_length" else 1
        return np.linalg.norm(soft) - (1 - l1_ratio) * alpha * scale

    # We use the brentq method to find the root, which requires a bracket
    # within which to find the root. We know that ``beta_zero_root`` will
    # be positive when alpha=0. In order to ensure that the upper limit
    # brackets the root, we increase the upper limit until
    # ``beta_zero_root`` returns a negative number for all groups
    def bracket_too_low(alpha):
        return any([beta_zero_root(alpha, group=grp) > 0 for grp in groups])

    upper_bracket_lim = 1e1
    while bracket_too_low(upper_bracket_lim):
        upper_bracket_lim *= 10

    min_alphas = np.array(
        [
            root_scalar(
                partial(beta_zero_root, group=grp),
                bracket=[0, upper_bracket_lim],
                method="brentq",
            ).root
            for grp in groups
        ]
    )

    alpha_max = np.max(min_alphas) * 1.2

    # Test feature sparsity just to make sure we're on the right side of the root
    while (  # pragma: no cover
        model(
            groups=groups,
            alpha=alpha_max,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            scale_l2_by=scale_l2_by,
        )
        .fit(X, y)
        .chosen_features_.size
        > 0
    ):
        alpha_max *= 1.2  # pragma: no cover

    if alpha_max <= np.finfo(float).resolution:
        alphas = np.empty(n_alphas)
        alphas.fill(np.finfo(float).resolution)
        return alphas

    return np.logspace(np.log10(alpha_max * eps), np.log10(alpha_max), num=n_alphas)[
        ::-1
    ]


def sgl_path(
    X,
    y,
    l1_ratio=0.5,
    groups=None,
    scale_l2_by="group_length",
    eps=1e-3,
    n_alphas=100,
    alphas=None,
    Xy=None,
    normalize=False,
    copy_X=True,
    verbose=False,
    return_n_iter=False,
    check_input=True,
    **params,
):
    """
    Compute sparse group lasso path.

    We use the previous solution as the initial guess for subsequent alpha values

    Parameters
    ----------
    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        Training data. Pass directly as Fortran-contiguous data to avoid
        unnecessary memory duplication.

    y : {array-like, sparse matrix} of shape (n_samples,)
        Target values.

    l1_ratio : float, default=0.5
        Number between 0 and 1 passed to SGL estimator (scaling between the
        group lasso and lasso penalties). ``l1_ratio=1`` corresponds to the
        Lasso.

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    eps : float, default=1e-3
        Length of the path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``.

    n_alphas : int, default=100
        Number of alphas along the regularization path.

    alphas : ndarray, default=None
        List of alphas where to compute the models.
        If None alphas are set automatically.

    Xy : array-like of shape (n_features,), default=None
        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
        prevent train/test leakage by ensuring the ``Xy`` is precomputed
        using only training data.

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : bool, default=True
        If ``True``, X will be copied; else, it may be overwritten.

    verbose : bool or int, default=False
        Amount of verbosity.

    check_input : bool, default=True
        Skip input validation checks, assuming there are handled by the
        caller when check_input=False.

    **params : kwargs
        Keyword arguments passed to the SGL estimator

    Returns
    -------
    coefs : ndarray of shape (n_features, n_alphas) or (n_features + 1, n_alphas)
        List of coefficients for the Logistic Regression model. If
        fit_intercept is set to True then the first dimension will be
        n_features + 1, where the last item represents the intercept.

    alphas : ndarray of shape (n_alphas,)
        The alphas along the path where models are computed.

    n_iters : array of shape (n_alphas,)
        Actual number of iteration for each alpha.

    See Also
    --------
    SGL
    SGLCV
    """
    # We expect X and y to be already Fortran ordered when bypassing
    # checks
    if check_input:
        X = check_array(
            X,
            accept_sparse=False,
            dtype=[np.float64, np.float32],
            order="F",
            copy=copy_X,
        )
        y = check_array(
            y,
            accept_sparse=False,
            dtype=X.dtype.type,
            order="F",
            copy=False,
            ensure_2d=False,
        )
        if Xy is not None:
            # Xy should be a 1d contiguous array
            Xy = check_array(
                Xy, dtype=X.dtype.type, order="C", copy=False, ensure_2d=False
            )
        groups = check_groups(groups, X, allow_overlap=False, fit_intercept=False)

    _, n_features = X.shape

    fit_intercept = params.get("fit_intercept", True)

    if alphas is None:
        alphas = _alpha_grid(
            X=X,
            y=y,
            Xy=Xy,
            groups=groups,
            scale_l2_by=scale_l2_by,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            eps=eps,
            n_alphas=n_alphas,
            normalize=normalize,
            copy_X=copy_X,
        )
    else:
        alphas = np.sort(alphas)[::-1]  # make sure alphas are properly ordered

    n_alphas = len(alphas)
    tol = params.get("tol", 1e-7)
    max_iter = params.get("max_iter", 1000)
    loss = params.get("loss", "squared_loss")
    n_iters = np.empty((n_alphas,), dtype=int)

    if fit_intercept:
        coefs = np.empty((n_features + 1, n_alphas), dtype=X.dtype)
    else:
        coefs = np.empty((n_features, n_alphas), dtype=X.dtype)

    model = SGL(
        l1_ratio=l1_ratio,
        alpha=alphas[0],
        groups=groups,
        scale_l2_by=scale_l2_by,
        fit_intercept=fit_intercept,
        max_iter=max_iter,
        tol=tol,
        warm_start=True,
        verbose=False,
        suppress_solver_warnings=True,
        include_solver_trace=False,
    )

    if verbose and verbose == 1:
        alpha_sequence = tqdm(alphas, desc="Reg path", total=n_alphas)
    else:
        alpha_sequence = alphas

    for i, alpha in enumerate(alpha_sequence):
        model.set_params(alpha=alpha)
        model.fit(X, y, loss=loss)

        if fit_intercept:
            coefs[..., i] = np.concatenate([model.coef_, [model.intercept_]])
        else:
            coefs[..., i] = model.coef_

        n_iters[i] = model.n_iter_

        if verbose:
            if verbose > 2:
                print(model)
            elif verbose > 1:
                print("Path: %03i out of %03i" % (i, n_alphas))

    return coefs, alphas, n_iters


def sgl_scoring_path(
    X,
    y,
    train,
    test,
    l1_ratio=0.5,
    groups=None,
    scale_l2_by="group_length",
    eps=1e-3,
    n_alphas=100,
    alphas=None,
    Xy=None,
    normalize=False,
    copy_X=True,
    verbose=False,
    check_input=True,
    scoring=None,
    **params,
):
    """Return the scores for the models computed by 'path'.

    Parameters
    ----------
    X : {array-like, sparse matrix} of shape (n_samples, n_features)
        Training data.

    y : array-like of shape (n_samples,) or (n_samples, n_targets)
        Target values.

    train : list of indices
        The indices of the train set.

    test : list of indices
        The indices of the test set.

    l1_ratio : float, default=0.5
        Number between 0 and 1 passed to SGL estimator (scaling between the
        group lasso and lasso penalties). ``l1_ratio=1`` corresponds to the
        Lasso.

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    eps : float, default=1e-3
        Length of the path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``.

    n_alphas : int, default=100
        Number of alphas along the regularization path.

    alphas : ndarray, default=None
        List of alphas where to compute the models.
        If None alphas are set automatically.

    Xy : array-like of shape (n_features,), default=None
        Xy = np.dot(X.T, y) that can be precomputed. If supplying ``Xy``,
        prevent train/test leakage by ensuring the ``Xy`` is precomputed
        using only training data.

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    copy_X : bool, default=True
        If ``True``, X will be copied; else, it may be overwritten.

    verbose : bool or int, default=False
        Amount of verbosity.

    check_input : bool, default=True
        Skip input validation checks, assuming there are handled by the
        caller when check_input=False.

    scoring : callable, default=None
        A string (see sklearn model evaluation documentation) or a scorer
        callable object / function with signature ``scorer(estimator, X, y)``.
        For a list of scoring functions that can be used, look at
        `sklearn.metrics`. The default scoring option used is accuracy_score.

    **params : kwargs
        Keyword arguments passed to the SGL estimator

    Returns
    -------
    coefs : ndarray of shape (n_features, n_alphas) or (n_features + 1, n_alphas)
        List of coefficients for the SGL model. If fit_intercept is set to
        True then the first dimension will be n_features + 1, where the last
        item represents the intercept.

    alphas : ndarray
        Grid of alphas used for cross-validation.

    scores : ndarray of shape (n_alphas,)
        Scores obtained for each alpha.

    n_iter : ndarray of shape(n_alphas,)
        Actual number of iteration for each alpha.
    """
    if Xy is not None:
        logger.warning(
            "You supplied the `Xy` parameter. Remember to ensure "
            "that Xy is computed from the training data alone."
        )

    X_train = X[train]
    y_train = y[train]
    X_test = X[test]
    y_test = y[test]

    coefs, alphas, n_iter = sgl_path(
        X_train,
        y_train,
        l1_ratio=l1_ratio,
        groups=groups,
        scale_l2_by=scale_l2_by,
        eps=eps,
        n_alphas=n_alphas,
        alphas=alphas,
        Xy=Xy,
        normalize=normalize,
        copy_X=copy_X,
        verbose=verbose,
        check_input=False,
        **params,
    )
    del X_train, y_train

    fit_intercept = params.get("fit_intercept", True)
    max_iter = params.get("max_iter", 1000)
    tol = params.get("tol", 1e-7)

    model = SGL(
        l1_ratio=l1_ratio,
        groups=groups,
        scale_l2_by=scale_l2_by,
        fit_intercept=fit_intercept,
        max_iter=max_iter,
        tol=tol,
        warm_start=True,
        verbose=False,
        suppress_solver_warnings=True,
        include_solver_trace=False,
    )

    if scoring is None:
        scoring = "neg_mean_squared_error"

    model.is_fitted_ = True

    scores = list()
    scoring = get_scorer(scoring)
    for w in coefs.T:
        if fit_intercept:
            model.coef_ = w[:-1]
            model.intercept_ = w[-1]
        else:
            model.coef_ = w
            model.intercept_ = 0.0

        scores.append(scoring(model, X_test, y_test))

    return coefs, alphas, np.array(scores), n_iter


class SGLCV(LinearModel, RegressorMixin, TransformerMixin):
    """Iterative SGL model fitting along a regularization path.

    See the scikit-learn glossary entry for `cross-validation estimator
    <https://scikit-learn.org/stable/glossary.html#term-cross-validation-estimator>`_

    Parameters
    ----------
    l1_ratio : float or list of float, default=1.0
        float between 0 and 1 passed to SGL (scaling between group lasso and
        lasso penalties). For ``l1_ratio = 0`` the penalty is the group lasso
        penalty. For ``l1_ratio = 1`` it is the lasso penalty. For ``0 <
        l1_ratio < 1``, the penalty is a combination of group lasso and
        lasso. This parameter can be a list, in which case the different
        values are tested by cross-validation and the one giving the best
        prediction score is used. Note that a good choice of list of values
        will depend on the problem. For problems where we expect strong
        overall sparsity and would like to encourage grouping, put more
        values close to 1 (i.e. Lasso). In contrast, if we expect strong
        group-wise sparsity, but only mild sparsity within groups, put more
        values close to 0 (i.e. group lasso).

    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group. We set groups in ``__init__`` so
        that it can be reused in model selection and CV routines.

    scale_l2_by : ["group_length", None], default="group_length"
        Scaling technique for the group-wise L2 penalty.
        By default, ``scale_l2_by="group_length`` and the L2 penalty is
        scaled by the square root of the group length so that each variable
        has the same effect on the penalty. This may not be appropriate for
        one-hot encoded features and ``scale_l2_by=None`` would be more
        appropriate for that case. ``scale_l2_by=None`` will also reproduce
        ElasticNet results when all features belong to one group.

    eps : float, default=1e-3
        Length of the path. ``eps=1e-3`` means that
        ``alpha_min / alpha_max = 1e-3``.

    n_alphas : int, default=100
        Number of alphas along the regularization path, used for each l1_ratio.

    alphas : ndarray, default=None
        List of alphas where to compute the models.
        If None alphas are set automatically

    fit_intercept : bool, default=True
        whether to calculate the intercept for this model. If set
        to false, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    normalize : bool, default=False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`sklearn:sklearn.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    max_iter : int, default=1000
        The maximum number of iterations

    tol : float, default=1e-7
        Stopping criterion. Convergence tolerance for the ``copt`` proximal gradient solver

    cv : int, cross-validation generator or iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - int, to specify the number of folds.
        - an sklearn `CV splitter <https://scikit-learn.org/stable/glossary.html#term-cv-splitter>`_,
        - An iterable yielding (train, test) splits as arrays of indices.

        For int/None inputs, :class:`sklearn:sklearn.model_selection.KFold` is used.

        Refer to the :ref:`scikit-learn User Guide
        <sklearn:cross_validation>` for the various cross-validation
        strategies that can be used here.

    copy_X : bool, default=True
        If ``True``, X will be copied; else, it may be overwritten.

    verbose : bool or int, default=0
        Amount of verbosity.

    n_jobs : int, default=None
        Number of CPUs to use during the cross validation.
        ``None`` means 1 unless in a :obj:`joblib:joblib.parallel_backend` context.
        ``-1`` means using all processors.

    tuning_strategy : ["grid", "bayes"], default="grid"
        Hyperparameter tuning strategy to use. If ``tuning_strategy == "grid"``,
        then evaluate all parameter points on the ``l1_ratio`` and ``alphas`` grid,
        using warm start to evaluate different ``alpha`` values along the
        regularization path. If ``tuning_strategy == "bayes"``, then a fixed
        number of parameter settings is sampled using ``skopt.BayesSearchCV``.
        The fixed number of settings is set by ``n_bayes_iter``. The
        ``l1_ratio`` setting is sampled uniformly from the minimum and maximum
        of the input ``l1_ratio`` parameter. The ``alpha`` setting is sampled
        log-uniformly either from the maximum and minumum of the input ``alphas``
        parameter, if provided or from ``eps`` * max_alpha to max_alpha where
        max_alpha is a conservative estimate of the maximum alpha for which the
        solution coefficients are non-trivial.

    n_bayes_iter : int, default=50
        Number of parameter settings that are sampled if using Bayes search
        for hyperparameter optimization. ``n_bayes_iter`` trades off runtime
        vs quality of the solution. Consider increasing ``n_bayes_points`` if
        you want to try more parameter settings in parallel.

    n_bayes_points : int, default=1
        Number of parameter settings to sample in parallel if using Bayes
        search for hyperparameter optimization. If this does not align with
        ``n_bayes_iter``, the last iteration will sample fewer points.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    suppress_solver_warnings : bool, default=True
        If True, suppress warnings from BayesSearchCV when the objective is
        evaluated at the same point multiple times. Setting this to False,
        may be useful for debugging.

    Attributes
    ----------
    alpha_ : float
        The amount of penalization chosen by cross validation

    l1_ratio_ : float
        The compromise between l1 and l2 penalization chosen by
        cross validation

    coef_ : ndarray of shape (n_features,) or (n_targets, n_features)
        Parameter vector (w in the cost function formula),

    intercept_ : float or ndarray of shape (n_targets, n_features)
        Independent term in the decision function.

    scoring_path_ : ndarray of shape (n_l1_ratio, n_alpha, n_folds)
        Mean square error for the test set on each fold, varying l1_ratio and
        alpha.

    alphas_ : ndarray of shape (n_alphas,) or (n_l1_ratio, n_alphas)
        The grid of alphas used for fitting, for each l1_ratio.

    n_iter_ : int
        number of iterations run by the proximal gradient descent solver to
        reach the specified tolerance for the optimal alpha.

    bayes_optimizer_ : skopt.BayesSearchCV instance or None
        The BayesSearchCV instance used for hyperparameter optimization if
        ``tuning_strategy == "bayes"``. If ``tuning_strategy == "grid"``,
        then this attribute is None.

    See Also
    --------
    sgl_path
    SGL
    """

    def __init__(
        self,
        l1_ratio=1.0,
        groups=None,
        scale_l2_by="group_length",
        eps=1e-3,
        n_alphas=100,
        alphas=None,
        fit_intercept=True,
        normalize=False,
        max_iter=1000,
        tol=1e-7,
        copy_X=True,
        cv=None,
        verbose=False,
        n_jobs=None,
        tuning_strategy="grid",
        n_bayes_iter=50,
        n_bayes_points=1,
        random_state=None,
        suppress_solver_warnings=True,
    ):
        self.l1_ratio = l1_ratio
        self.groups = groups
        self.scale_l2_by = scale_l2_by
        self.eps = eps
        self.n_alphas = n_alphas
        self.alphas = alphas
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.max_iter = max_iter
        self.tol = tol
        self.copy_X = copy_X
        self.cv = cv
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.tuning_strategy = tuning_strategy
        self.n_bayes_iter = n_bayes_iter
        self.n_bayes_points = n_bayes_points
        self.random_state = random_state
        self.suppress_solver_warnings = suppress_solver_warnings

    def fit(self, X, y):
        """Fit sparse group lasso linear model.

        Fit is on grid of alphas and best alpha estimated by cross-validation.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data. Pass directly as Fortran-contiguous data
            to avoid unnecessary memory duplication. If y is mono-output,
            X can be sparse.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values
        """
        if self.tuning_strategy.lower() not in ["grid", "bayes"]:
            raise ValueError(
                "`tuning_strategy` must be either 'grid' or 'bayes'; got "
                "{0} instead.".format(self.tuning_strategy)
            )

        # This makes sure that there is no duplication in memory.
        # Dealing right with copy_X is important in the following:
        # Multiple functions touch X and subsamples of X and can induce a
        # lot of duplication of memory
        copy_X = self.copy_X and self.fit_intercept

        check_y_params = dict(
            copy=False, dtype=[np.float64, np.float32], ensure_2d=False
        )

        if isinstance(X, np.ndarray) or sparse.isspmatrix(X):
            # Keep a reference to X
            reference_to_old_X = X
            # Let us not impose fortran ordering so far: it is
            # not useful for the cross-validation loop and will be done
            # by the model fitting itself

            # Need to validate separately here.
            # We can't pass multi_ouput=True because that would allow y to be
            # csr. We also want to allow y to be 64 or 32 but check_X_y only
            # allows to convert for 64.
            check_X_params = dict(
                accept_sparse=False, dtype=[np.float64, np.float32], copy=False
            )
            X, y = self._validate_data(
                X, y, validate_separately=(check_X_params, check_y_params)
            )
            if not np.may_share_memory(reference_to_old_X, X):
                # X has been copied
                copy_X = False
            del reference_to_old_X
        else:
            # Need to validate separately here.
            # We can't pass multi_ouput=True because that would allow y to be
            # csr. We also want to allow y to be 64 or 32 but check_X_y only
            # allows to convert for 64.
            check_X_params = dict(
                accept_sparse=False,
                dtype=[np.float64, np.float32],
                order="F",
                copy=copy_X,
            )
            X, y = self._validate_data(
                X, y, validate_separately=(check_X_params, check_y_params)
            )
            copy_X = False

        model = SGL()
        y = column_or_1d(y, warn=True)

        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "X and y have inconsistent dimensions (%d != %d)"
                % (X.shape[0], y.shape[0])
            )

        groups = check_groups(self.groups, X, allow_overlap=False, fit_intercept=False)

        # All SGLCV parameters except "cv" and "n_jobs" are acceptable
        path_params = self.get_params()
        path_params.pop("cv", None)
        path_params.pop("n_jobs", None)
        path_params.pop("tuning_strategy", None)
        path_params.pop("n_bayes_iter", None)
        path_params.pop("n_bayes_points", None)
        path_params.pop("random_state", None)

        l1_ratios = np.atleast_1d(path_params["l1_ratio"])
        # For the first path, we need to set l1_ratio
        path_params["l1_ratio"] = l1_ratios[0]

        alphas = self.alphas
        n_l1_ratio = len(l1_ratios)
        if alphas is None:
            alphas = [
                _alpha_grid(
                    X=X,
                    y=y,
                    groups=groups,
                    scale_l2_by=self.scale_l2_by,
                    l1_ratio=l1_ratio,
                    fit_intercept=self.fit_intercept,
                    eps=self.eps,
                    n_alphas=self.n_alphas,
                    normalize=self.normalize,
                    copy_X=self.copy_X,
                )
                for l1_ratio in l1_ratios
            ]
        else:
            # Making sure alphas is properly ordered.
            alphas = np.tile(np.sort(alphas)[::-1], (n_l1_ratio, 1))

        # We want n_alphas to be the number of alphas used for each l1_ratio.
        n_alphas = len(alphas[0])
        path_params.update({"n_alphas": n_alphas})

        path_params["copy_X"] = copy_X
        # We are not computing in parallel, we can modify X
        # inplace in the folds
        if effective_n_jobs(self.n_jobs) > 1:
            path_params["copy_X"] = False

        if isinstance(self.verbose, int):  # pragma: no cover
            path_params["verbose"] = self.verbose - 1

        # init cross-validation generator
        cv = check_cv(self.cv)

        if self.tuning_strategy == "grid":
            # Compute path for all folds and compute MSE to get the best alpha
            folds = list(cv.split(X, y))
            best_score = -np.inf

            path_params.pop("cv", None)
            path_params.pop("n_jobs", None)
            path_params.pop("alphas", None)
            path_params.pop("l1_ratio", None)
            path_params.update({"groups": groups})

            # We do a double for loop folded in one, in order to be able to
            # iterate in parallel on l1_ratio and folds
            jobs = (
                delayed(sgl_scoring_path)(
                    X=X,
                    y=y,
                    train=train,
                    test=test,
                    l1_ratio=this_l1_ratio,
                    alphas=this_alphas,
                    Xy=None,
                    **path_params,
                )
                for this_l1_ratio, this_alphas in zip(l1_ratios, alphas)
                for train, test in folds
            )

            if isinstance(self.verbose, int):  # pragma: no cover
                parallel_verbosity = self.verbose - 2
                if parallel_verbosity < 0:
                    parallel_verbosity = 0
            else:  # pragma: no cover
                parallel_verbosity = self.verbose

            score_paths = Parallel(
                n_jobs=self.n_jobs,
                verbose=parallel_verbosity,
                **_joblib_parallel_args(prefer="threads"),
            )(jobs)

            coefs_paths, alphas_paths, scores, n_iters = zip(*score_paths)
            scores = np.reshape(scores, (n_l1_ratio, len(folds), -1))
            alphas_paths = np.reshape(alphas_paths, (n_l1_ratio, len(folds), -1))
            n_iters = np.reshape(n_iters, (n_l1_ratio, len(folds), -1))
            coefs_paths = np.reshape(
                coefs_paths, (n_l1_ratio, len(folds), -1, n_alphas)
            )

            mean_score = np.mean(scores, axis=1)
            self.scoring_path_ = np.squeeze(np.moveaxis(scores, 2, 1))

            for l1_ratio, l1_alphas, score_alphas in zip(l1_ratios, alphas, mean_score):
                i_best_alpha = np.argmax(score_alphas)
                this_best_score = score_alphas[i_best_alpha]
                if this_best_score > best_score:
                    best_alpha = l1_alphas[i_best_alpha]
                    best_l1_ratio = l1_ratio
                    best_score = this_best_score

            self.l1_ratio_ = best_l1_ratio
            self.alpha_ = best_alpha

            if self.alphas is None:
                self.alphas_ = np.asarray(alphas)
                if n_l1_ratio == 1:
                    self.alphas_ = self.alphas_[0]
            # Remove duplicate alphas in case alphas is provided.
            else:
                self.alphas_ = np.asarray(alphas[0])

            # Refit the model with the parameters selected
            common_params = {
                name: value
                for name, value in self.get_params().items()
                if name in model.get_params()
            }

            model.set_params(**common_params)
            model.alpha = best_alpha
            model.l1_ratio = best_l1_ratio
            model.copy_X = copy_X

            model.fit(X, y)

            self.coef_ = model.coef_
            self.intercept_ = model.intercept_
            self.n_iter_ = model.n_iter_
            self.bayes_optimizer_ = None
            self.is_fitted_ = True
        else:
            # Set the model with the common input params
            common_params = {
                name: value
                for name, value in self.get_params().items()
                if name in model.get_params()
            }
            model.set_params(**common_params)
            model.copy_X = copy_X

            search_spaces = {"alpha": (np.min(alphas), np.max(alphas), "log-uniform")}

            l1_min = np.min(self.l1_ratio)
            l1_max = np.max(self.l1_ratio)

            if l1_min < l1_max:
                search_spaces["l1_ratio"] = (
                    np.min(self.l1_ratio),
                    np.max(self.l1_ratio),
                    "uniform",
                )

            self.bayes_optimizer_ = BayesSearchCV(
                model,
                search_spaces,
                n_iter=self.n_bayes_iter,
                cv=cv,
                n_jobs=self.n_jobs,
                n_points=self.n_bayes_points,
                verbose=self.verbose,
                random_state=self.random_state,
                return_train_score=True,
                scoring="neg_mean_squared_error",
                error_score=-np.inf,
            )

            if self.suppress_solver_warnings:
                ctx_mgr = warnings.catch_warnings()
            else:
                ctx_mgr = contextlib.suppress()

            with ctx_mgr:
                # If n_bayes_points > 1 the objective may be evaluated at the
                # same point multiple times. This is okay and we give the user
                # the choice as to whether or not to see these warnings. The
                # default is to suppress them.
                if self.suppress_solver_warnings:
                    warnings.filterwarnings("ignore", category=UserWarning)

                self.bayes_optimizer_.fit(X, y)

            self.alpha_ = self.bayes_optimizer_.best_estimator_.alpha
            self.l1_ratio_ = self.bayes_optimizer_.best_estimator_.l1_ratio
            self.coef_ = self.bayes_optimizer_.best_estimator_.coef_
            self.intercept_ = self.bayes_optimizer_.best_estimator_.intercept_
            self.n_iter_ = self.bayes_optimizer_.best_estimator_.n_iter_
            self.is_fitted_ = True
            self.scoring_path_ = None
            param_alpha = self.bayes_optimizer_.cv_results_["param__alpha"]
            self.alphas_ = np.sort(param_alpha)[::-1]

        return self

    @property
    def chosen_features_(self):
        """Return index array of chosen features."""
        return np.nonzero(self.coef_)[0]

    @property
    def sparsity_mask_(self):
        """Return boolean array indicating which features survived regularization."""
        return self.coef_ != 0

    def like_nonzero_mask_(self, rtol=1e-8):
        """Return boolean array indicating which features are zero or close to zero.

        Parameters
        ----------
        rtol : float
            Relative tolerance. Any features that are larger in magnitude
            than ``rtol`` times the mean coefficient value are considered
            nonzero-like.
        """
        mean_abs_coef = abs(self.coef_.mean())
        return np.abs(self.coef_) > rtol * mean_abs_coef

    @property
    def chosen_groups_(self):
        """Return set of the group IDs that survived regularization."""
        if self.groups is not None:
            group_mask = [
                bool(set(grp).intersection(set(self.chosen_features_)))
                for grp in self.groups
            ]
            return np.nonzero(group_mask)[0]
        else:
            return self.chosen_features_

    def transform(self, X):
        """Remove columns corresponding to zeroed-out coefficients."""
        # Check is fit had been called
        check_is_fitted(self, "is_fitted_")

        # Input validation
        X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        if X.shape[1] != self.coef_.size:
            raise ValueError("Shape of input is different from what was seen in `fit`")

        return X[:, self.sparsity_mask_]

    def _more_tags(self):  # pylint: disable=no-self-use
        return {"multioutput": False, "requires_y": True}



"""
======================================================
Logistic Sparse Group Lasso for grouped sparse signals
======================================================

Estimates a Sparse Group Lasso logistic regression model on a simulated
sparse signal. The estimated important features are compared with the
ground-truth.

"""
import numpy as np
from matplotlib import pyplot as plt
from groupyr import LogisticSGLCV
from groupyr.datasets import make_group_classification

X, y, groups, idx = make_group_classification(
    n_samples=100,
    n_groups=20,
    n_informative_groups=3,
    n_features_per_group=20,
    n_informative_per_group=10,
    n_redundant_per_group=0,
    n_repeated_per_group=0,
    n_classes=2,
    scale=100,
    useful_indices=True,
    random_state=1729,
)

_, n_features = X.shape

model = LogisticSGLCV(
    groups=groups, l1_ratio=[0.80, 0.90], n_alphas=40, tol=1e-3, eps=1e-2, cv=3
).fit(X, y)

plt.plot(
    np.arange(n_features),
    model.coef_,
    marker="o",
    mfc="black",
    mec="none",
    ms=3,
    mew=0,
    ls="",
    label="coefficients",
)

plt.plot(
    np.arange(n_features)[idx],
    model.coef_[idx],
    marker="o",
    mfc="none",
    mec="green",
    ms=5,
    mew=3,
    ls="",
    label="informative features",
)

plt.title("Estimated coefficients with ground truth imporant features highlighted")

plt.legend(loc="best")
plt.xlabel("Feature index")
plt.ylabel("Coefs")

plt.show()

print("Indices of ground-truth informative features:")
print(np.where(idx)[0])

print("Indices of non-zero estimated coefs:")
print(model.chosen_features_)



from __future__ import absolute_import, division, print_function

import numpy as np
import pytest

from groupyr.datasets import make_group_classification, make_group_regression

from sklearn.utils._testing import assert_array_almost_equal
from sklearn.utils._testing import assert_raises


@pytest.mark.parametrize("shuffle", [True, False])
@pytest.mark.parametrize("return_idx", [True, False])
def test_make_group_sparse_classification(shuffle, return_idx):
    weights = [0.10, 0.25]

    clf_res = make_group_classification(
        n_samples=100,
        n_groups=10,
        n_features_per_group=8,
        n_informative_groups=5,
        n_informative_per_group=2,
        n_redundant_per_group=1,
        n_repeated_per_group=1,
        n_classes=3,
        n_clusters_per_class=1,
        hypercube=False,
        shift=None,
        scale=None,
        weights=weights,
        random_state=0,
        shuffle=shuffle,
        useful_indices=return_idx,
    )

    if return_idx:
        X, y, groups, idx = clf_res
        assert np.sum(idx) == 20  # nosec
    else:
        X, y, groups = clf_res

    assert weights == [0.1, 0.25]  # nosec
    assert X.shape == (100, 80), "X shape mismatch"  # nosec
    assert y.shape == (100,), "y shape mismatch"  # nosec
    assert len(groups) == 10, "groups shape mismatch"  # nosec
    assert np.unique(y).shape == (3,), "Unexpected number of classes"  # nosec
    assert sum(y == 0) == 10, "Unexpected number of samples in class #0"  # nosec
    assert sum(y == 1) == 25, "Unexpected number of samples in class #1"  # nosec
    assert sum(y == 2) == 65, "Unexpected number of samples in class #2"  # nosec

    # Test for n_features > 30
    X, y, groups = make_group_classification(
        n_samples=2000,
        n_groups=11,
        n_features_per_group=3,
        n_informative_groups=11,
        n_informative_per_group=3,
        n_redundant_per_group=0,
        n_repeated_per_group=0,
        n_clusters_per_class=1,
        n_classes=3,
        hypercube=True,
        flip_y=0,
        shift=None,
        scale=0.5,
        random_state=0,
    )

    assert X.shape == (2000, 33), "X shape mismatch"  # nosec
    assert y.shape == (2000,), "y shape mismatch"  # nosec
    assert len(groups) == 11, "groups shape mismatch"  # nosec

    assert (  # nosec
        np.unique(X.view([("", X.dtype)] * X.shape[1]))
        .view(X.dtype)
        .reshape(-1, X.shape[1])
        .shape[0]
        == 2000
    ), "Unexpected number of unique rows"

    assert_raises(
        ValueError,
        make_group_classification,
        n_groups=20,
        n_features_per_group=2,
        n_informative_groups=20,
        n_redundant_per_group=1,
        n_repeated_per_group=2,
    )
    assert_raises(ValueError, make_group_classification, weights=weights, n_classes=5)


def test_make_group_regression():
    X, y, groups, coefs = make_group_regression(
        n_samples=100,
        n_groups=2,
        n_informative_groups=1,
        n_features_per_group=5,
        n_informative_per_group=2,
        effective_rank=5,
        coef=True,
        random_state=0,
    )

    assert X.shape == (100, 10), "X shape mismatch"  # nosec
    assert y.shape == (100,), "y shape mismatch"  # nosec
    assert len(groups) == 2, "groups shape mismatch"  # nosec
    assert coefs.shape == (10,), "coef shape mismatch"  # nosec
    assert sum(coefs != 0.0) == 2, "Unexpected number of informative features"  # nosec

    X, y, groups, coefs = make_group_regression(
        n_samples=2000,
        n_groups=2,
        n_informative_groups=2,
        n_features_per_group=5,
        n_informative_per_group=3,
        effective_rank=10,
        noise=1.0,
        coef=True,
        shuffle=True,
        random_state=0,
    )

    assert X.shape == (2000, 10), "X shape mismatch"  # nosec
    assert y.shape == (2000,), "y shape mismatch"  # nosec
    assert len(groups) == 2, "groups shape mismatch"  # nosec
    assert coefs.shape == (10,), "coef shape mismatch"  # nosec
    assert sum(coefs != 0.0) == 6, "Unexpected number of informative features"  # nosec
    assert (  # nosec
        np.unique(X.view([("", X.dtype)] * X.shape[1]))
        .view(X.dtype)
        .reshape(-1, X.shape[1])
        .shape[0]
        == 2000
    ), "Unexpected number of unique rows"

    # Test that y ~= np.dot(X, coefs) + bias + N(0, 1.0).
    assert_array_almost_equal(np.std(y - np.dot(X, coefs)), 1.0, decimal=1)

    # Test with small number of features.
    X, y, groups = make_group_regression(
        n_samples=100,
        n_groups=1,
        n_informative_groups=1,
        n_features_per_group=1,
        n_informative_per_group=1,
    )  # n_informative=3
    assert X.shape == (100, 1)  # nosec


@pytest.mark.parametrize(
    "make_dataset", [make_group_classification, make_group_regression]
)
@pytest.mark.parametrize("shuffle", [True, False])
def test_random_state_returns_same_data(make_dataset, shuffle):
    X0, y0, groups0 = make_dataset(random_state=1729, shuffle=shuffle)
    X1, y1, groups1 = make_dataset(random_state=1729, shuffle=shuffle)

    assert_array_almost_equal(X0, X1)
    assert_array_almost_equal(y0, y1)
    assert_array_almost_equal(groups0, groups1)

    if make_dataset == make_group_regression:
        kwargs = {"coef": True, "shuffle": shuffle}
    elif make_dataset == make_group_classification:
        kwargs = {"useful_indices": True, "shuffle": shuffle}

    X0, y0, groups0, coef_idx0 = make_dataset(random_state=1729, **kwargs)
    X1, y1, groups1, coef_idx1 = make_dataset(random_state=1729, **kwargs)

    assert_array_almost_equal(X0, X1)
    assert_array_almost_equal(y0, y1)
    assert_array_almost_equal(groups0, groups1)
    assert_array_almost_equal(coef_idx0, coef_idx1)



"""
=============================================================
Rudimentary performance comparison of groupyr and group-lasso
=============================================================

This example performs a rudimentary performance comparison of the groupyr and
group-lasso packages for various problem sizes. Speedup is defined as the
ratio of time to fit a `group-lasso.GroupLasso` model to the time to fit a
`groupyr.SGL` model.

In addition to `groupyr`, you will need to install the `group-lasso` package
using
```
pip install group-lasso
```
"""
import groupyr as gpr
import matplotlib.pyplot as plt
import numpy as np

from group_lasso import GroupLasso
from sklearn.model_selection import train_test_split
from timeit import timeit
from tqdm.auto import tqdm

n_features = [5, 10, 25, 50, 75, 100, 150, 200, 250]
speedup = []

for n_features_per_group in tqdm(n_features):
    X, y, groups, coef = gpr.datasets.make_group_regression(
        n_samples=400,
        n_groups=50,
        n_informative_groups=5,
        n_features_per_group=n_features_per_group,
        n_informative_per_group=int(0.8 * n_features_per_group),
        noise=200,
        coef=True,
        random_state=10,
        shuffle=True,
    )

    gl_groups = np.concatenate(
        [[idx] * len(grp) for idx, grp in enumerate(groups)]
    ).reshape([-1, 1])

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.25, random_state=10
    )

    sgl = gpr.SGL(groups=groups, l1_ratio=0.5, alpha=1.0, max_iter=1000, tol=1e-3)

    gl = GroupLasso(
        groups=gl_groups,
        group_reg=0.5,
        l1_reg=0.5,
        frobenius_lipschitz=True,
        scale_reg="group_size",
        n_iter=1000,
        tol=1e-3,
        supress_warning=True,
    )

    gpr_time = timeit("sgl.fit(X_train, y_train)", globals=globals(), number=3)
    gl_time = timeit("gl.fit(X_train, y_train)", globals=globals(), number=3)

    speedup.append(gpr_time / gl_time)

fig, ax = plt.subplots(1, 1, figsize=(8, 5))

_ = ax.plot(np.array(n_features) * 50, 1 / np.array(speedup))
_ = ax.set_xlabel(r"Number of features", fontsize=16)
_ = ax.set_ylabel(r"$t$(group-lasso) / $t$(groupyr)", fontsize=16)
_ = ax.set_title(
    "\n".join([r"Rudimentary performance comparison", r"of groupyr and group-lasso"]),
    fontsize=16,
)

fig.savefig("./groupyr_speedup.pdf", bbox_inches="tight")



import numpy as np
import pytest

from groupyr import SGL, SGLCV, LogisticSGL, LogisticSGLCV
from groupyr.datasets import make_group_regression, make_group_classification
from groupyr.sgl import _alpha_grid

from skopt import BayesSearchCV
from sklearn.linear_model.tests.test_coordinate_descent import build_dataset
from sklearn.model_selection import GridSearchCV, StratifiedKFold
from sklearn.utils._testing import assert_almost_equal, assert_array_almost_equal


def test_sgl_input_validation():
    X = [[0], [0], [0]]
    y = [0, 0, 0]

    with pytest.raises(ValueError):
        SGL(l1_ratio=1.0, alpha=0.1, warm_start="error").fit(X, y)

    with pytest.raises(ValueError):
        SGL(l1_ratio=1.0, alpha=0.1).fit(X, y, loss="error")

    with pytest.raises(ValueError):
        SGL(l1_ratio=100, alpha=0.1).fit(X, y)

    with pytest.raises(ValueError):
        SGL(l1_ratio=0.5, alpha=0.1, scale_l2_by="error").fit(X, y)

    with pytest.raises(ValueError):
        _alpha_grid(X, y, l1_ratio=0.5, scale_l2_by="error")


@pytest.mark.parametrize("Estimator", [SGL, SGLCV, LogisticSGL, LogisticSGLCV])
def test_sgl_masks(Estimator):
    groups = [np.arange(5), np.arange(5, 10)]
    model = Estimator(groups=groups)
    coefs_0 = np.concatenate([np.ones(5), np.zeros(5)])
    model.coef_ = coefs_0
    assert_array_almost_equal(model.chosen_features_, np.arange(5))
    assert_array_almost_equal(model.sparsity_mask_, coefs_0 != 0)
    assert_array_almost_equal(model.chosen_groups_, np.array([0]))

    model.groups = None
    assert_array_almost_equal(model.chosen_groups_, np.arange(5))

    coefs_1 = np.concatenate([np.ones(5), np.ones(5) * 1e-7])
    model.coef_ = coefs_1
    assert_array_almost_equal(model.like_nonzero_mask_(rtol=2.1e-7), coefs_0 != 0)
    assert all(model.like_nonzero_mask_(rtol=1e-8))  # nosec


@pytest.mark.parametrize("suppress_warnings", [True, False])
def test_sgl_zero(suppress_warnings):
    # Check that SGL can handle zero data without crashing
    X = [[0], [0], [0]]
    y = [0, 0, 0]
    clf = SGL(l1_ratio=1.0, alpha=0.1, suppress_solver_warnings=suppress_warnings).fit(
        X, y
    )
    pred = clf.predict([[1], [2], [3]])
    assert_array_almost_equal(clf.coef_, [0])
    assert_array_almost_equal(pred, [0, 0, 0])


# When l1_ratio=1, SGL should behave like the lasso. These next tests
# replicate the unit testing for sklearn.linear_model.Lasso
@pytest.mark.parametrize("loss", ["squared_loss", "huber"])
def test_sgl_toy(loss):
    # Test on a toy example for various values of l1_ratio.
    # When validating this against glmnet notice that glmnet divides it
    # against n_obs.

    X = [[-1], [0], [1]]
    y = [-1, 0, 1]  # just a straight line
    T = [[2], [3], [4]]  # test sample

    tol = 1e-6

    clf = SGL(alpha=1e-8, tol=tol)
    clf.fit(X, y, loss=loss)
    pred = clf.predict(T)
    assert_array_almost_equal(clf.coef_, [1])
    assert_array_almost_equal(pred, [2, 3, 4])

    clf = SGL(alpha=0.1, tol=tol)
    clf.fit(X, y, loss=loss)
    pred = clf.predict(T)
    assert_array_almost_equal(clf.coef_, [0.85])
    assert_array_almost_equal(pred, [1.7, 2.55, 3.4])

    clf = SGL(alpha=0.5, tol=tol)
    clf.fit(X, y, loss=loss)
    pred = clf.predict(T)
    assert_array_almost_equal(clf.coef_, [0.25])
    assert_array_almost_equal(pred, [0.5, 0.75, 1.0])

    clf = SGL(alpha=1, tol=tol)
    clf.fit(X, y, loss=loss)
    pred = clf.predict(T)
    assert_array_almost_equal(clf.coef_, [0.0])
    assert_array_almost_equal(pred, [0, 0, 0])


@pytest.mark.parametrize("fit_intercept", [True, False])
def test_warm_start(fit_intercept):
    X, y, _, _ = build_dataset()
    max_iter = 5
    sgl = SGL(
        l1_ratio=0.5,
        alpha=0.1,
        warm_start=True,
        max_iter=max_iter,
        fit_intercept=fit_intercept,
        include_solver_trace=True,
    )
    sgl.fit(X, y)
    tr_0 = sgl.solver_trace_
    sgl.fit(X, y)
    tr_1 = sgl.solver_trace_

    sgl = SGL(
        l1_ratio=0.5,
        alpha=0.1,
        max_iter=max_iter + 1,
        fit_intercept=fit_intercept,
        include_solver_trace=True,
    )
    sgl.fit(X, y)
    tr_2 = sgl.solver_trace_

    fx_cold_start = np.array(tr_2.trace_fx)
    fx_warm_start = np.array(tr_0.trace_fx + tr_1.trace_fx)[: fx_cold_start.size]

    assert_array_almost_equal(fx_warm_start, fx_cold_start)


@pytest.mark.parametrize("tuning_strategy", ["grid", "bayes"])
def test_sgl_cv(tuning_strategy):
    X, y, X_test, y_test = build_dataset()
    max_iter = 500
    clf = SGLCV(
        n_alphas=10,
        eps=1e-3,
        max_iter=max_iter,
        cv=3,
        tuning_strategy=tuning_strategy,
        random_state=42,
        n_bayes_iter=10,
    ).fit(X, y)
    assert_almost_equal(clf.alpha_, 0.056, 2)
    assert clf.score(X_test, y_test) > 0.99  # nosec

    if tuning_strategy == "grid":
        alphas = np.copy(clf.alphas_)
        np.random.default_rng().shuffle(alphas)
        clf2 = SGLCV(
            alphas=alphas,
            max_iter=max_iter,
            cv=3,
            n_jobs=2,
            tuning_strategy=tuning_strategy,
        ).fit(X, y)
        assert_array_almost_equal(clf.alphas_, clf2.alphas_)

        # Make sure that this attr is set to None:
        assert clf2.bayes_optimizer_ is None

    else:
        clf = SGLCV(
            l1_ratio=[0.95, 1.0],
            n_alphas=10,
            eps=1e-3,
            max_iter=max_iter,
            cv=3,
            tuning_strategy=tuning_strategy,
            random_state=42,
            n_bayes_iter=20,
        ).fit(X, y)
        assert clf.score(X_test, y_test) > 0.98  # nosec
        assert_almost_equal(clf.alpha_, 0.06, 2)


@pytest.mark.parametrize("execution_number", range(5))
@pytest.mark.parametrize("l1_ratio", np.random.default_rng(seed=1729).uniform(size=4))
@pytest.mark.parametrize("compute_Xy", [True, "ndim2", False])
def test_alpha_grid_starts_at_zero(execution_number, l1_ratio, compute_Xy):
    X, y, groups = make_group_regression()

    if compute_Xy is True:
        Xy = X.T.dot(y)
    elif compute_Xy == "ndim2":
        Xy = X.T.dot(y)[:, np.newaxis]
    else:
        Xy = None

    agrid = _alpha_grid(X, y, Xy=Xy, groups=groups, l1_ratio=l1_ratio, n_alphas=10)
    model = SGL(groups=groups, l1_ratio=l1_ratio, alpha=agrid[0]).fit(X, y)

    assert model.chosen_features_.size == 0  # nosec


def test_alpha_grid_float_resolution():
    X = np.zeros((3, 3))
    y = [1, 2, 3]
    agrid = _alpha_grid(X, y, l1_ratio=0.5, n_alphas=5)
    assert_array_almost_equal(agrid, np.ones(5) * np.finfo(float).resolution)


def test_sglcv_value_errors():
    X = [[0], [0], [0]]
    y = [0, 0, 0]

    with pytest.raises(ValueError):
        SGLCV().fit(X, y + [0])

    with pytest.raises(ValueError):
        SGLCV(tuning_strategy="error").fit(X, y)

    with pytest.raises(ValueError):
        LogisticSGLCV(tuning_strategy="error").fit(X, y)


def test_LogisticSGLCV_GridSearchCV():
    # make sure LogisticSGLCV gives same best params (l1_ratio and alpha) as
    # GridSearchCV
    X, y, groups = make_group_classification(random_state=42)

    cv = StratifiedKFold(3)
    l1_ratios = np.linspace(0, 1, 3)
    alphas = np.logspace(-4, 4, 3)

    clf_cv = LogisticSGLCV(alphas=alphas, l1_ratio=l1_ratios, groups=groups, cv=cv)
    clf_cv.fit(X, y)

    param_grid = {"alpha": alphas, "l1_ratio": l1_ratios}
    clf = LogisticSGL(groups=groups)
    gs = GridSearchCV(clf, param_grid, cv=cv)
    gs.fit(X, y)

    assert gs.best_params_["l1_ratio"] == clf_cv.l1_ratio_
    assert gs.best_params_["alpha"] == clf_cv.alpha_


def test_LogisticSGLCV_BayesSearchCV():
    # make sure LogisticSGLCV gives same best params (l1_ratio and alpha) as
    # BayesSearchCV
    X, y, groups = make_group_classification(random_state=42)
    cv = StratifiedKFold(3)

    l1_ratios = np.linspace(0, 1, 3)
    alphas = np.logspace(-4, 4, 3)
    clf_cv = LogisticSGLCV(
        alphas=alphas,
        l1_ratio=l1_ratios,
        groups=groups,
        cv=cv,
        tuning_strategy="bayes",
        n_bayes_iter=10,
        random_state=42,
    )
    clf_cv.fit(X, y)

    search_spaces = {
        "alpha": (np.min(alphas), np.max(alphas), "log-uniform"),
        "l1_ratio": (np.min(l1_ratios), np.max(l1_ratios), "uniform"),
    }
    clf = LogisticSGL(groups=groups)
    gs = BayesSearchCV(clf, search_spaces, cv=cv, random_state=42, n_iter=10)
    gs.fit(X, y)

    assert gs.best_params_["l1_ratio"] == clf_cv.l1_ratio_
    assert gs.best_params_["alpha"] == clf_cv.alpha_



"""Utility functions for SGL-based estimators."""
import numpy as np


def check_groups(groups, X, allow_overlap=False, fit_intercept=False, kwarg_name="X"):
    """Validate group indices.

    Verify that all features in ``X`` are accounted for in groups,
    that all groups refer to features that actually exist in ``X``,
    and, if ``allow_overlap=False``, that all groups are distinct.

    Parameters
    ----------
    groups : list of numpy.ndarray
        list of arrays of non-overlapping indices for each group. For
        example, if nine features are grouped into equal contiguous groups of
        three, then groups would be ``[array([0, 1, 2]), array([3, 4, 5]),
        array([6, 7, 8])]``. If the feature matrix contains a bias or
        intercept feature, do not include it as a group. If None, all
        features will belong to one group.

    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        The training input samples. If ``X`` includes a bias or intercept
        feature, it must be in the last column and ``fit_intercept`` should
        be ``True``.

    allow_overlap : bool, default=False
        If True, allow groups to overlap. i.e. each feature may belong to
        multiple groups

    fit_intercept : bool, default=False
        If True, assume that the last column of the feature matrix
        corresponds to the bias or intercept.

    kwarg_name : str, default="X"
        The keyword argument name used to customize error messages. Defaults to
        "X" as this function is most commonly used to check feature matrices
        during fit, transform, and predict functions.

    Returns
    -------
    groups : list of numpy.ndarray
        The validated groups.
    """
    n_features = X.shape[-1]

    if fit_intercept:
        n_features -= 1

    if groups is None:
        # If no groups provided, put all features in one group
        return [np.arange(n_features)]

    all_indices = np.concatenate(groups)

    if set(all_indices) < set(range(n_features)):
        raise ValueError(
            "Some features are unaccounted for in groups; Columns "
            "{0} are absent from groups.".format(
                set(range(n_features)) - set(all_indices)
            )
        )

    if set(all_indices) > set(range(n_features)):
        raise ValueError(
            "There are feature indices in groups that exceed the dimensions "
            "of {0}; {0} has {1} features but groups refers to indices {2}".format(
                kwarg_name, n_features, set(all_indices) - set(range(n_features))
            )
        )

    if not allow_overlap:
        _, counts = np.unique(all_indices, return_counts=True)
        if set(counts) != {1}:
            raise ValueError("Overlapping groups detected.")

    return tuple(groups)
